"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define StormEurope class.
"""

__all__ = ['StormEurope']

import bz2
import logging
from pathlib import Path
import numpy as np
import xarray as xr
import pandas as pd
import matplotlib.pyplot as plt
from scipy import sparse
import datetime as dt

from climada.util.config import CONFIG
from climada.hazard.base import Hazard
from climada.hazard.centroids.centr import Centroids
from climada.hazard.tag import Tag as TagHazard
from climada.util.files_handler import get_file_names
from climada.util.dates_times import (datetime64_to_ordinal,
                                      last_year,
                                      first_year,
                                      date_to_str
                                      )
from climada.util.dwd_icon_loader import (download_icon_centroids_file,
                                          download_icon_grib,
                                          delete_icon_grib,
                                          )

LOGGER = logging.getLogger(__name__)

HAZ_TYPE = 'WS'
"""Hazard type acronym for Winter Storm"""

N_PROB_EVENTS = 5 * 6
"""Number of events per historic event in probabilistic dataset"""

FORECAST_DIR = CONFIG.hazard.storm_europe.forecast_dir.str()


class StormEurope(Hazard):
    """A hazard set containing european winter storm events. Historic storm
    events can be downloaded at http://wisc.climate.copernicus.eu/ and read
    with `from_footprints`. Weather forecasts can be automatically downloaded from
    https://opendata.dwd.de/ and read with from_icon_grib(). Weather forecast
    from the COSMO-Consortium http://www.cosmo-model.org/ can be read with
    from_cosmoe_file().

    Attributes
    ----------
    ssi_wisc : np.array, float
        Storm Severity Index (SSI) as recorded in
        the footprint files; apparently not reproducible from the footprint
        values only.
    ssi : np.array, float
        SSI as set by set_ssi; uses the Dawkins
        definition by default.
    """

    intensity_thres = 14.7
    """Intensity threshold for storage in m/s; same as used by WISC SSI calculations."""

    vars_opt = Hazard.vars_opt.union({'ssi_wisc', 'ssi', 'ssi_full_area'})
    """Name of the variables that aren't need to compute the impact."""

    def __init__(self):
        """Calls the Hazard init dunder. Sets unit to 'm/s'."""
        Hazard.__init__(self, HAZ_TYPE)
        self.units = 'm/s'
        self.ssi = np.array([], float)
        self.ssi_wisc = np.array([], float)
        self.ssi_full_area = np.array([], float)

    def read_footprints(self, *args, **kwargs):
        """This function is deprecated, use StormEurope.from_footprints instead."""
        LOGGER.warning("The use of StormEurope.read_footprints is deprecated."
                       "Use StormEurope.from_footprints instead.")
        self.__dict__ = StormEurope.from_footprints(*args, **kwargs).__dict__

    @classmethod
    def from_footprints(cls, path, description=None, ref_raster=None, centroids=None,
                        files_omit='fp_era20c_1990012515_701_0.nc', combine_threshold=None,
                        intensity_thres=None):
        """Create new StormEurope object from WISC footprints.

        Assumes that all footprints have the same coordinates as the first file listed/first
        file in dir.

        Parameters
        ----------
        path : str, list(str)
            A location in the filesystem. Either a
            path to a single netCDF WISC footprint, or a folder
            containing only footprints, or a globbing pattern to one or
            more footprints.
        description : str, optional
            description of the events, defaults
            to 'WISC historical hazard set'
        ref_raster : str, optional
            Reference netCDF file from which to
            construct a new barebones Centroids instance. Defaults to
            the first file in path.
        centroids : Centroids, optional
            A Centroids struct, overriding
            ref_raster
        files_omit : str, list(str), optional
            List of files to omit;
            defaults to one duplicate storm present in the WISC set as
            of 2018-09-10.
        combine_threshold : int, optional
            threshold for combining events
            in number of days. if the difference of the dates (self.date)
            of two events is smaller or equal to this threshold, the two
            events are combined into one.
            Default is None, Advised for WISC is 2
        intensity_thres : float, optional
            Intensity threshold for storage in m/s. Default: class attribute
            StormEurope.intensity_thres (same as used by WISC SSI calculations)

        Returns
        -------
        haz : StormEurope
            StormEurope object with data from WISC footprints.
        """
        intensity_thres = cls.intensity_thres if intensity_thres is None else intensity_thres
        file_names = get_file_names(path)

        if ref_raster is not None and centroids is not None:
            LOGGER.warning('Overriding ref_raster with centroids')

        if centroids is not None:
            pass
        elif ref_raster is not None:
            centroids = cls._centroids_from_nc(ref_raster)
        elif ref_raster is None:
            centroids = cls._centroids_from_nc(file_names[0])

        if isinstance(files_omit, str):
            files_omit = [files_omit]

        LOGGER.info('Commencing to iterate over netCDF files.')

        haz = cls()
        for file_name in file_names:
            if any(fo in file_name for fo in files_omit):
                LOGGER.info("Omitting file %s", file_name)
                continue
            new_haz = cls._read_one_nc(file_name, centroids, intensity_thres)
            if new_haz is not None:
                haz.append(new_haz)

        haz.event_id = np.arange(1, len(haz.event_id) + 1)
        haz.frequency = np.divide(
            np.ones_like(haz.date),
            np.max([(last_year(haz.date) - first_year(haz.date)), 1])
        )

        haz.tag = TagHazard(
            HAZ_TYPE, 'Hazard set not saved by default',
            description='WISC historical hazard set.'
        )
        if description is not None:
            haz.tag.description = description

        if combine_threshold is not None:
            LOGGER.info('Combining events with small difference in date.')
            difference_date = np.diff(haz.date)
            for event_id_i in haz.event_id[
                    np.append(difference_date <= combine_threshold, False)]:
                event_ids = [event_id_i, event_id_i + 1]
                haz._combine_events(event_ids)
        return haz

    @staticmethod
    def _read_one_nc(file_name, centroids, intensity_thres):
        """Read a single WISC footprint. Assumes a time dimension of length 1.
        Omits a footprint if another file with the same timestamp has already
        been read.

        Parameters
        ----------
        file_name : str
            Absolute or relative path to *.nc
        centroids : Centroids
            Centr. instance that matches the
            coordinates used in the *.nc, only validated by size.
        intensity_thres : float
            Intensity threshold for storage in m/s.

        Returns
        -------
        new_haz : StormEurope
            Hazard instance for one single storm.
        """
        ncdf = xr.open_dataset(file_name)

        if centroids.size != (ncdf.sizes['latitude'] * ncdf.sizes['longitude']):
            ncdf.close()
            LOGGER.warning(('Centroids size doesn\'t match NCDF dimensions. '
                            'Omitting file %s.'), file_name)
            return None

        # xarray does not penalise repeated assignments, see
        # http://xarray.pydata.org/en/stable/data-structures.html
        stacked = ncdf.max_wind_gust.stack(
            intensity=('latitude', 'longitude', 'time')
        )
        stacked = stacked.where(stacked > intensity_thres)
        stacked = stacked.fillna(0)

        # fill in values from netCDF
        new_haz = StormEurope()
        new_haz.event_name = [ncdf.storm_name]
        new_haz.date = np.array([datetime64_to_ordinal(ncdf.time.data[0])])
        new_haz.intensity = sparse.csr_matrix(stacked)
        new_haz.ssi_wisc = np.array([float(ncdf.ssi)])

        # fill in default values
        new_haz.centroids = centroids
        new_haz.event_id = np.array([1])
        new_haz.frequency = np.array([1])
        new_haz.fraction = new_haz.intensity.copy().tocsr()
        new_haz.fraction.data.fill(1)
        new_haz.orig = np.array([True])

        ncdf.close()
        return new_haz

    def read_cosmoe_file(self, *args, **kwargs):
        """This function is deprecated, use StormEurope.from_cosmoe_file instead."""
        LOGGER.warning("The use of StormEurope.read_cosmoe_file is deprecated."
                       "Use StormEurope.from_cosmoe_file instead.")
        self.__dict__ = StormEurope.from_cosmoe_file(*args, **kwargs).__dict__

    @classmethod
    def from_cosmoe_file(cls, fp_file, run_datetime, event_date=None,
                         model_name='COSMO-2E', description=None, intensity_thres=None):
        """Create a new StormEurope object with gust footprint from weather forecast.

        The funciton is designed for the COSMO ensemble model used by
        the COSMO Consortium http://www.cosmo-model.org/ and postprocessed to
        an netcdf file using fieldextra. One event is one full day in UTC.
        Works for MeteoSwiss model output of
        COSMO-1E (11 members, resolution 1.1 km, forecast period 33-45 hours)
        COSMO-2E (21 members, resolution 2.2 km, forecast period 5 days)

        The frequency of each event is informed by their probability in the ensemble
        forecast and is equal to 1/11 or 1/21 for COSMO-1E or COSMO-2E, respectively.

        Parameters
        ----------
        fp_file : str
            string directing to one netcdf file
        run_datetime : datetime
            The starting timepoint of the forecast run
            of the cosmo model
        event_date : datetime, optional
            one day within the forecast
            period, only this day (00H-24H) will be included in the hazard
        model_name : str,optional
            provide the name of the COSMO model,
            for the description (e.g., 'COSMO-1E', 'COSMO-2E')
        description : str, optional
            description of the events, defaults
            to a combination of model_name and run_datetime
        intensity_thres : float, optional
            Intensity threshold for storage in m/s. Default: class attribute
            StormEurope.intensity_thres (same as used by WISC SSI calculations)

        Returns
        -------
        haz : StormEurope
            StormEurope object with data from COSMO ensemble file.
        """
        intensity_thres = cls.intensity_thres if intensity_thres is None else intensity_thres

        haz = cls()
        # create centroids
        haz.centroids = cls._centroids_from_nc(fp_file)

        # read intensity from file
        ncdf = xr.open_dataset(fp_file)
        ncdf = ncdf.assign_coords(date=('time',ncdf["time"].dt.floor("D").values))

        if event_date:
            try:
                stacked = ncdf.sel(
                    time=event_date.strftime('%Y-%m-%d')
                    ).groupby('date').max().stack(intensity=('y_1', 'x_1'))
            except KeyError:
                raise ValueError('Extraction of date and coordinates failed. '
                                 'This is most likely because '
                                 'the selected event_date {} is not contained'
                                 ' in the weather forecast selected by '
                                 'fp_file {}. Please adjust event_date'
                                 ' or fp_file.'.format(
                                     event_date.strftime('%Y-%m-%d'),
                                     fp_file))
            considered_dates = np.datetime64(event_date)
        else:
            time_covered_step = ncdf['time'].diff('time')
            time_covered_day = time_covered_step.groupby('date').sum()
            # forecast run should cover at least 18 hours of a day
            considered_dates_bool = time_covered_day >= np.timedelta64(18,'h')
            stacked = ncdf.groupby('date'
                                    ).max().sel(date=considered_dates_bool
                                                ).stack(intensity=('y_1',
                                                                   'x_1'))
            considered_dates = stacked['date'].values
        stacked = stacked.stack(date_ensemble=('date', 'epsd_1'))
        stacked = stacked.where(stacked.VMAX_10M > intensity_thres)
        stacked = stacked.fillna(0)

        # fill in values from netCDF
        haz.intensity = sparse.csr_matrix(stacked.VMAX_10M.T)
        haz.event_id = np.arange(stacked.date_ensemble.size)+1


        # fill in default values
        haz.units = 'm/s'
        haz.fraction = haz.intensity.copy().tocsr()
        haz.fraction.data.fill(1)
        haz.orig = np.ones_like(haz.event_id)*False
        haz.orig[(stacked.epsd_1 == 0).values] = True
        haz.date = np.repeat(
            np.array(datetime64_to_ordinal(considered_dates)),
            np.unique(ncdf.epsd_1).size
            )
        haz.event_name = [date_i + '_ens' + str(ens_i)
                           for date_i, ens_i in zip(date_to_str(haz.date),
                                                    stacked.epsd_1.values+1)
                           ]
        haz.frequency = np.divide(
                np.ones_like(haz.event_id),
                np.unique(ncdf.epsd_1).size)
        if not description:
            description = (model_name +
                           ' weather forecast windfield ' +
                           'for run startet at ' +
                           run_datetime.strftime('%Y%m%d%H'))

        haz.tag = TagHazard(
                HAZ_TYPE, 'Hazard set not saved, too large to pickle',
                description=description
            )
        # close netcdf file
        ncdf.close()
        haz.check()
        return haz

    def read_icon_grib(self, *args, **kwargs):
        """This function is deprecated, use StormEurope.from_icon_grib instead."""
        LOGGER.warning("The use of StormEurope.read_icon_grib is deprecated."
                       "Use StormEurope.from_icon_grib instead.")
        self.__dict__ = StormEurope.from_icon_grib(*args, **kwargs).__dict__

    @classmethod
    def from_icon_grib(cls, run_datetime, event_date=None, model_name='icon-eu-eps',
                       description=None, grib_dir=None, delete_raw_data=True,
                       intensity_thres=None):
        """Create new StormEurope object from DWD icon weather forecast footprints.

        New files are available for 24 hours on
        https://opendata.dwd.de, old files can be processed if they are
        already stored in grib_dir.
        One event is one full day in UTC. Current setup works for runs
        starting at 00H and 12H. Otherwise the aggregation is inaccurate,
        because of the given file structure with 1-hour, 3-hour and
        6-hour maxima provided.

        The frequency for one event is 1/(number of ensemble members)

        Parameters
        ----------
        run_datetime : datetime
            The starting timepoint of the forecast run
            of the icon model
        event_date : datetime, optional
            one day within the forecast
            period, only this day (00H-24H) will be included in the hazard
        model_name : str,optional
            select the name of the icon model to
            be downloaded. Must match the url on https://opendata.dwd.de
            (see download_icon_grib for further info)
        description : str, optional
            description of the events, defaults
            to a combination of model_name and run_datetime
        grib_dir : str, optional
            path to folder, where grib files are
            or should be stored
        delete_raw_data : bool,optional
            select if downloaded raw data in
            .grib.bz2 file format should be stored on the computer or
            removed
        intensity_thres : float, optional
            Intensity threshold for storage in m/s. Default: class attribute
            StormEurope.intensity_thres (same as used by WISC SSI calculations)

        Returns
        -------
        haz : StormEurope
            StormEurope object with data from DWD icon weather forecast footprints.
        """
        intensity_thres = cls.intensity_thres if intensity_thres is None else intensity_thres

        haz = cls()
        if not (run_datetime.hour == 0 or run_datetime.hour == 12):
            LOGGER.warning('The event definition is inaccuratly implemented '+
                           'for starting times, which are not 00H or 12H.')
        # download files, if they don't already exist
        file_names = download_icon_grib(
            run_datetime, model_name=model_name, download_dir=grib_dir)

        # create centroids
        nc_centroids_file = download_icon_centroids_file(model_name, grib_dir)
        haz.centroids = haz._centroids_from_nc(nc_centroids_file)

        # read intensity from files
        for ind_i, file_i in enumerate(file_names):
            gripfile_path_i = Path(file_i[:-4])
            with open(file_i, 'rb') as source, open(gripfile_path_i, 'wb') as dest:
                dest.write(bz2.decompress(source.read()))
            ds_i = xr.open_dataset(gripfile_path_i, engine='cfgrib')
            if ind_i == 0:
                stacked = ds_i
            else:
                stacked = xr.concat([stacked,ds_i], 'valid_time')

        # create intensity matrix with max for each full day
        stacked = stacked.assign_coords(
            date=('valid_time', stacked["valid_time"].dt.floor("D").values))
        if event_date:
            try:
                stacked = stacked.sel(
                    valid_time=event_date.strftime('%Y-%m-%d')).groupby('date').max()
            except KeyError:
                raise ValueError('Extraction of date and coordinates failed. '
                                 'This is most likely because '
                                 'the selected event_date {} is not contained'
                                 ' in the weather forecast selected by '
                                 'run_datetime {}. Please adjust event_date'
                                 ' or run_datetime.'.format(
                                     event_date.strftime('%Y-%m-%d'),
                                     run_datetime.strftime('%Y-%m-%d %H:%M')))

            considered_dates = np.datetime64(event_date)
        else:
            time_covered_step = stacked['valid_time'].diff('valid_time')
            time_covered_day = time_covered_step.groupby('date').sum()
            # forecast run should cover at least 18 hours of a day
            considered_dates_bool = time_covered_day >= np.timedelta64(18,'h')
            stacked = stacked.groupby('date').max().sel(date=considered_dates_bool)
            considered_dates = stacked['date'].values
        stacked = stacked.stack(date_ensemble=('date', 'number'))
        stacked = stacked.where(stacked > intensity_thres)
        stacked = stacked.fillna(0)


        # fill in values from netCDF
        haz.intensity = sparse.csr_matrix(stacked.gust.T)
        haz.event_id = np.arange(stacked.date_ensemble.size)+1

        # fill in default values
        haz.units = 'm/s'
        haz.fraction = haz.intensity.copy().tocsr()
        haz.fraction.data.fill(1)
        haz.orig = np.ones_like(haz.event_id)*False
        haz.orig[(stacked.number == 1).values] = True

        haz.date = np.repeat(
            np.array(datetime64_to_ordinal(considered_dates)),
            np.unique(stacked.number).size
            )
        haz.event_name = [date_i + '_ens' + str(ens_i)
                          for date_i, ens_i in zip(date_to_str(haz.date), stacked.number.values)]
        haz.frequency = np.divide(
                np.ones_like(haz.event_id),
                np.unique(stacked.number).size)
        if not description:
            description = ('icon weather forecast windfield ' +
                           'for run startet at ' +
                           run_datetime.strftime('%Y%m%d%H'))

        haz.tag = TagHazard(
            HAZ_TYPE, 'Hazard set not saved, too large to pickle',
            description=description
            )
        haz.check()

        # delete generated .grib2 and .4cc40.idx files
        for ind_i, file_i in enumerate(file_names):
            gripfile_path_i = Path(file_i[:-4])
            idxfile_path_i = next(gripfile_path_i.parent.glob(
                str(gripfile_path_i.name) + '.*.idx'))
            gripfile_path_i.unlink()
            idxfile_path_i.unlink()

        if delete_raw_data:
            #delete downloaded .bz2 files
            delete_icon_grib(run_datetime, model_name=model_name, download_dir=grib_dir)

        return haz

    @staticmethod
    def _centroids_from_nc(file_name):
        """Construct Centroids from the grid described by 'latitude' and
        'longitude' variables in a netCDF file.
        """
        LOGGER.info('Constructing centroids from %s', file_name)
        cent = Centroids()
        ncdf = xr.open_dataset(file_name)
        create_meshgrid = True
        if hasattr(ncdf, 'latitude'):
            lats = ncdf.latitude.data
            lons = ncdf.longitude.data
        elif hasattr(ncdf, 'lat'):
            lats = ncdf.lat.data
            lons = ncdf.lon.data
        elif hasattr(ncdf, 'lat_1'):
            if len(ncdf.lon_1.shape)>1 & \
                (ncdf.lon_1.shape == ncdf.lat_1.shape) \
                :
                lats = ncdf.lat_1.data.flatten()
                lons = ncdf.lon_1.data.flatten()
                create_meshgrid = False
            else:
                lats = ncdf.lat_1.data
                lons = ncdf.lon_1.data
        elif hasattr(ncdf, 'clat'):
            lats = ncdf.clat.data
            lons = ncdf.clon.data
            if ncdf.clat.attrs['units']=='radian':
                lats = np.rad2deg(lats)
                lons = np.rad2deg(lons)
            create_meshgrid = False
        else:
            raise AttributeError('netcdf file has no field named latitude or '
                                 'other know abrivation for coordinates.')
        ncdf.close()

        if create_meshgrid:
            lats, lons = np.array([np.repeat(lats, len(lons)),
                                   np.tile(lons, len(lats))])
        cent = Centroids.from_lat_lon(lats, lons)
        cent.set_area_pixel()
        cent.set_on_land()

        return cent

    def _combine_events(self, event_ids):
        """combine the intensities of two events using max and adjust event_id, event_name,
        date etc of the hazard

        the event_ids must be consecutive for the event_name field to behave correctly

        Parameters
        ----------
        event_ids : array
            two consecutive event ids
        """
        select_event_ids = np.isin(self.event_id, event_ids)
        select_other_events = np.invert(select_event_ids)
        intensity_tmp = self.intensity[select_event_ids, :].max(axis=0)
        self.intensity = self.intensity[select_other_events, :]
        self.intensity = sparse.vstack([self.intensity, sparse.csr_matrix(intensity_tmp)])
        self.event_id = np.append(self.event_id[select_other_events], self.event_id.max() + 1)
        self.date = np.append(self.date[select_other_events],
                              np.round(self.date[select_event_ids].mean()))
        name_2 = self.event_name.pop(np.where(select_event_ids)[0][1])
        name_1 = self.event_name.pop(np.where(select_event_ids)[0][0])
        self.event_name.append(name_1 + '_' + name_2)
        fraction_tmp = self.fraction[select_event_ids, :].max(axis=0)
        self.fraction = self.fraction[select_other_events, :]
        self.fraction = sparse.vstack([self.fraction, sparse.csr_matrix(fraction_tmp)])

        self.frequency = np.append(self.frequency[select_other_events],
                                   self.frequency[select_event_ids].mean())
        self.orig = np.append(self.orig[select_other_events],
                              self.orig[select_event_ids].max())
        if self.ssi_wisc.size > 0:
            self.ssi_wisc = np.append(self.ssi_wisc[select_other_events],
                                      np.nan)
        if self.ssi.size > 0:
            self.ssi = np.append(self.ssi[select_other_events],
                                 np.nan)
        if self.ssi_full_area.size > 0:
            self.ssi_full_area = np.append(self.ssi_full_area[select_other_events],
                                           np.nan)
        self.check()

    def calc_ssi(self, method='dawkins', intensity=None, on_land=True,
                 threshold=None, sel_cen=None):
        """Calculate the SSI, method must either be 'dawkins' or 'wisc_gust'.

        'dawkins', after Dawkins et al. (2016),
        doi:10.5194/nhess-16-1999-2016, matches the MATLAB version.
        ssi = sum_i(area_cell_i * intensity_cell_i^3)

        'wisc_gust', according to the WISC Tier 1 definition found at
        https://wisc.climate.copernicus.eu/wisc/#/help/products#tier1_section
        ssi = sum(area_on_land) * mean(intensity)^3

        In both definitions, only raster cells that are above the threshold are
        used in the computation.
        Note that this method does not reproduce self.ssi_wisc, presumably
        because the footprint only contains the maximum wind gusts instead of
        the sustained wind speeds over the 72 hour window. The deviation may
        also be due to differing definitions of what lies on land (i.e. Syria,
        Russia, Northern Africa and Greenland are exempt).

        Parameters
        ----------
        method : str
            Either 'dawkins' or 'wisc_gust'
        intensity : scipy.sparse.csr
            Intensity matrix; defaults to
            self.intensity
        on_land : bool
            Only calculate the SSI for areas on land,
            ignoring the intensities at sea. Defaults to true, whereas
            the MATLAB version did not.
        threshold : float, optional
            Intensity threshold used in index
            definition. Cannot be lower than the read-in value.
        sel_cen : np.array, bool
            A boolean vector selecting centroids.
            Takes precendence over on_land.

        Attributes
        ----------
        self.ssi_dawkins : np.array
            SSI per event
        """
        if intensity is not None:
            if not isinstance(intensity, sparse.csr_matrix):
                intensity = sparse.csr_matrix(intensity)
            else:
                pass
        else:
            intensity = self.intensity

        if threshold is not None:
            assert threshold >= self.intensity_thres, \
                'threshold cannot be below threshold upon read_footprint'
            intensity = intensity.multiply(intensity > threshold)
        else:
            intensity = intensity.multiply(intensity > self.intensity_thres)

        cent = self.centroids

        if sel_cen is not None:
            pass
        elif on_land is True:
            sel_cen = cent.on_land
        else:  # select all centroids
            sel_cen = np.ones_like(cent.area_pixel, dtype=bool)

        ssi = np.zeros(intensity.shape[0])

        if method == 'dawkins':
            area_c = cent.area_pixel / 1000 / 1000 * sel_cen
            for i, inten_i in enumerate(intensity):
                ssi_i = area_c * inten_i.power(3).todense().T
                # matrix crossproduct (row x column vector)
                ssi[i] = ssi_i.item(0)

        elif method == 'wisc_gust':
            for i, inten_i in enumerate(intensity[:, sel_cen]):
                area = np.sum(cent.area_pixel[inten_i.indices]) / 1000 / 1000
                inten_mean = np.mean(inten_i)
                ssi[i] = area * np.power(inten_mean, 3)

        return ssi

    def set_ssi(self, **kwargs):
        """Wrapper around calc_ssi for setting the self.ssi attribute.

        Parameters
        ----------
        kwargs :
            passed on to calc_ssi

        Attributes
        ----------
        ssi : np.array
            SSI per event
        """
        self.ssi = self.calc_ssi(**kwargs)

    def plot_ssi(self, full_area=False):
        """Plot the distribution of SSIs versus their cumulative exceedance
            frequencies, highlighting historical storms in red.

        Returns
        -------
        fig : matplotlib.figure.Figure
        ax : matplotlib.axes._subplots.AxesSubplot
        """
        if full_area:
            ssi = self.ssi_full_area
        else:
            ssi = self.ssi

        # data wrangling
        ssi_freq = pd.DataFrame({
            'ssi': ssi,
            'freq': self.frequency,
            'orig': self.orig,
        })
        ssi_freq = ssi_freq.sort_values('ssi', ascending=False)
        ssi_freq['freq_cum'] = np.cumsum(ssi_freq.freq)

        ssi_hist = ssi_freq.loc[ssi_freq.orig].copy()
        ssi_hist.freq = ssi_hist.freq * self.orig.size / self.orig.sum()
        ssi_hist['freq_cum'] = np.cumsum(ssi_hist.freq)

        # plotting
        fig, axs = plt.subplots()
        axs.plot(ssi_freq.freq_cum, ssi_freq.ssi, label='All Events')
        axs.scatter(ssi_hist.freq_cum, ssi_hist.ssi,
                    color='red', label='Historic Events')
        axs.legend()
        axs.set_xlabel('Exceedance Frequency [1/a]')
        axs.set_ylabel('Storm Severity Index')
        axs.ticklabel_format(axis='y', style='sci', scilimits=(0, 0))
        plt.show()

        return fig, axs

    def generate_prob_storms(self, reg_id=528, spatial_shift=4, ssi_args=None,
                             **kwargs):
        """Generates a new hazard set with one original and 29 probabilistic
        storms per historic storm. This represents a partial implementation of
        the Monte-Carlo method described in section 2.2 of Schwierz et al.
        (2010), doi:10.1007/s10584-009-9712-1.
        It omits the rotation of the storm footprints, as well as the pseudo-
        random alterations to the intensity.

        In a first step, the original intensity and five additional intensities
        are saved to an array. In a second step, those 6 possible intensity
        levels are shifted by n raster pixels into each direction (N/S/E/W).

        Caveats:
            - Memory safety is an issue; trial with the entire dataset resulted
              in 60GB of swap memory being used...
            - Can only use numeric region_id for country selection
            - Drops event names as provided by WISC

        Parameters
        ----------
        region_id : int, list of ints, or None
            iso_n3 code of the
            countries we want the generated hazard set to be returned for.
        spatial_shift : int
            amount of raster pixels to shift by
        ssi_args : dict
            A dictionary of arguments passed to calc_ssi
        kwargs :
            keyword arguments passed on to self._hist2prob()

        Returns
        -------
        new_haz : StormEurope
            A new hazard set for the given country.
            Centroid attributes are preserved. self.orig attribute is set
            to True for original storms (event_id ending in 00). Also
            contains a ssi_prob attribute,
        """
        # bool vector selecting the targeted centroids
        if reg_id is not None:
            if self.centroids.region_id.size == 0:
                self.centroids.set_region_id()
            if not isinstance(reg_id, list):
                reg_id = [reg_id]
            sel_cen = np.isin(self.centroids.region_id, reg_id)

        else:  # shifting truncates valid centroids
            sel_cen = np.zeros(self.centroids.shape, bool)
            sel_cen[
                spatial_shift:-spatial_shift,
                spatial_shift:-spatial_shift
            ] = True
            sel_cen = sel_cen.reshape(self.centroids.size)

        # init probabilistic array
        n_out = N_PROB_EVENTS * self.size
        intensity_prob = sparse.lil_matrix((n_out, np.count_nonzero(sel_cen)))
        ssi = np.zeros(n_out)

        LOGGER.info('Commencing probabilistic calculations')
        for index, intensity1d in enumerate(self.intensity):
            # indices for return matrix
            start = index * N_PROB_EVENTS
            end = (index + 1) * N_PROB_EVENTS

            intensity_prob[start:end, :], ssi[start:end] =\
                self._hist2prob(
                    intensity1d,
                    sel_cen,
                    spatial_shift,
                    ssi_args,
                    **kwargs)

        LOGGER.info('Generating new StormEurope instance')
        new_haz = StormEurope()
        new_haz.intensity = sparse.csr_matrix(intensity_prob)
        new_haz.ssi_full_area = ssi

        # don't use synthetic dates; just repeat the historic dates
        new_haz.date = np.repeat(self.date, N_PROB_EVENTS)

        # subsetting centroids
        new_haz.centroids = self.centroids.select(sel_cen=sel_cen)

        # construct new event ids
        base = np.repeat((self.event_id * 100), N_PROB_EVENTS)
        synth_id = np.tile(np.arange(N_PROB_EVENTS), self.size)
        new_haz.event_id = base + synth_id

        # frequency still based on the historic number of years
        new_haz.frequency = np.divide(np.repeat(self.frequency, N_PROB_EVENTS),
                                      N_PROB_EVENTS)

        new_haz.tag = TagHazard(
            HAZ_TYPE, 'Hazard set not saved by default',
            description='WISC probabilistic hazard set according to Schwierz et al.'
        )

        new_haz.fraction = new_haz.intensity.copy().tocsr()
        new_haz.fraction.data.fill(1)
        new_haz.orig = (new_haz.event_id % 100 == 0)

        new_haz.check()

        return new_haz

    def _hist2prob(self, intensity1d, sel_cen, spatial_shift, ssi_args=None,
                   power=1.15, scale=0.0225):
        """Internal function, intended to be called from generate_prob_storms.
        Generates six permutations based on one historical storm event, which
        it then moves around by spatial_shift gridpoints to the east, west, and
        north.

        Parameters
        ----------
        intensity1d : scipy.sparse.csr_matrix, 1 by n
            One historic event
        sel_cen : np.ndarray(dtype=bool)
            which centroids to return
        spatial_shift : int
            amount of raster cells to shift by
        power : float
            power to be applied elementwise
        scale : float
            weight of probabilistic component
        ssi_args : dict
            named arguments passed on to calc_ssi

        Returns
        -------
        intensity : np.array
            Synthetic intensities of shape (N_PROB_EVENTS, length(sel_cen))
        ssi : np.array
            SSI per synthetic event according to provided method.
        """
        if not ssi_args:
            ssi_args = {}
        shape_ndarray = tuple([N_PROB_EVENTS]) + self.centroids.shape

        # reshape to the raster that the data represents
        intensity2d = intensity1d.reshape(self.centroids.shape)

        # scipy.sparse.csr.csr_matrix elementwise methods (to avoid this:
        # https://github.com/ContinuumIO/anaconda-issues/issues/9129 )
        intensity2d_sqrt = intensity2d.power(1.0 / power).todense()
        intensity2d_pwr = intensity2d.power(power).todense()
        intensity2d = intensity2d.todense()

        # intermediary 3d array: (lat, lon, events)
        intensity3d_prob = np.ndarray(shape_ndarray)

        # the six variants of intensity transformation
        # 1. translation only
        intensity3d_prob[0] = intensity2d

        # 2. and 3. plusminus scaled sqrt
        intensity3d_prob[1] = intensity2d - (scale * intensity2d_sqrt)
        intensity3d_prob[2] = intensity2d + (scale * intensity2d_sqrt)

        # 4. and 5. plusminus scaled power
        intensity3d_prob[3] = intensity2d - (scale * intensity2d_pwr)
        intensity3d_prob[4] = intensity2d + (scale * intensity2d_pwr)

        # 6. minus scaled sqrt and pwr
        intensity3d_prob[5] = (intensity2d
                               - (0.5 * scale * intensity2d_pwr)
                               - (0.5 * scale * intensity2d_sqrt))

        # spatial shifts
        # northward
        intensity3d_prob[6:12, :-spatial_shift, :] = \
            intensity3d_prob[0:6, spatial_shift:, :]
        # southward
        intensity3d_prob[12:18, spatial_shift:, :] = \
            intensity3d_prob[0:6, :-spatial_shift, :]
        # eastward
        intensity3d_prob[18:24, :, spatial_shift:] = \
            intensity3d_prob[0:6, :, :-spatial_shift]
        # westward
        intensity3d_prob[24:30, :, :-spatial_shift] = \
            intensity3d_prob[0:6, :, spatial_shift:]

        intensity_out = intensity3d_prob.reshape(
            N_PROB_EVENTS,
            np.prod(self.centroids.shape)
        )

        ssi = self.calc_ssi(intensity=intensity_out, **ssi_args)

        return intensity_out[:, sel_cen], ssi


def generate_WS_forecast_hazard(run_datetime = dt.datetime.today().replace(hour=0,
                                                                           minute=0,
                                                                           second=0,
                                                                           microsecond=0),
                                event_date = (dt.datetime.today().replace(hour=0,
                                                                          minute=0,
                                                                          second=0,
                                                                          microsecond=0)
                                              + dt.timedelta(days=2)),
                                haz_model = 'icon-eu-eps',
                                haz_raw_storage = None,
                                save_haz = True):
    """ use the initialization time (run_datetime), the date of the event and
    specify the forecast model (haz_model) to generate a Hazard from forecast
    data either by download or through reading from existing file.

    Parameters
    ----------
    run_datetime: datetime.datetime, optional
        The starting timepoint of the forecast run
        of the icon model, defaults to today 00:00 hours
    event_date: datetime.datetime, optional
        one day within the forecast
        period, only this day (00H-24H) will be included in the hazard,
        defaults to the day after tomorrow
    haz_model: str, optional
        select the name of the model to
        be used: one of ['icon-eu-eps', 'icon-d2-eps', 'cosmo1e_file',
                         'cosmo2e_file'],
        default is 'icon-eu-eps'.
    haz_raw_storage: str, optional
        path to folder, where netcdf files
        are stored, mendatory when haz_model is 'cosmo1e_file' or
        'cosmo2e_file'. the string must contain one set of curly brakets,
        used by the function str.format to include
        run_datetime.strftime('%y%m%d%H').
        Default None resolves to "cosmoe_forecast_{}_vmax.nc" in
        CONFIG.hazard.storm_europe.forecast_dir
    save_haz: bool, optional
        flag if resulting hazard should be saved in
        CONFIG.hazard.storm_europe.forecast_dir, default is True.
    Returns
    -------
    Hazard
        hazard
    str
        haz_model
    datetime.datetime
        run_datetime
    datetime.datetime
        event_date
    """
    if (haz_model == 'cosmo1e_file'
        or
        haz_model == 'cosmo2e_file'):
        if haz_model == 'cosmo1e_file':
            haz_model='C1E'
            full_model_name_temp = 'COSMO-1E'
        if haz_model == 'cosmo2e_file':
            haz_model='C2E'
            full_model_name_temp = 'COSMO-2E'
        haz_file_name = Path(FORECAST_DIR) / (HAZ_TYPE +
                                              '_' +
                                              haz_model +
                                              '_run' +
                                              run_datetime.strftime('%Y%m%d%H') +
                                              '_event' +
                                              event_date.strftime('%Y%m%d')
                                              +
                                              '.hdf5')
        if haz_file_name.exists():
            LOGGER.info('Loading hazard from ' +
                        str(haz_file_name) +
                        '.')
            hazard = StormEurope.from_hdf5(haz_file_name)
        else:
            LOGGER.info('Generating ' +
                        haz_model +
                        ' hazard.')
            if not haz_raw_storage:
                haz_raw_storage = Path(FORECAST_DIR) / "cosmoe_forecast_{}_vmax.nc"
            fp_file = Path(
                str(haz_raw_storage).format(
                    run_datetime.strftime('%y%m%d%H')
                    )
                )
            hazard = StormEurope.from_cosmoe_file(
                fp_file,
                event_date=event_date,
                run_datetime=run_datetime,
                model_name=full_model_name_temp
                )
            if save_haz:
                hazard.write_hdf5(haz_file_name)
    elif (haz_model == 'icon-eu-eps'
          or
          haz_model == 'icon-d2-eps'):
        if haz_model == 'icon-eu-eps':
            full_model_name_temp = haz_model
            haz_model='IEE'
        if haz_model == 'icon-d2-eps':
            full_model_name_temp = haz_model
            haz_model='IDE'
        haz_file_name = Path(FORECAST_DIR) / (HAZ_TYPE +
                                              '_' +
                                              haz_model +
                                              '_run' +
                                              run_datetime.strftime('%Y%m%d%H') +
                                              '_event' +
                                              event_date.strftime('%Y%m%d')
                                              +
                                              '.hdf5')
        if haz_file_name.exists():
            LOGGER.info('Loading hazard from ' +
                        str(haz_file_name) +
                        '.')
            hazard = StormEurope.from_hdf5(haz_file_name)
        else:
            LOGGER.info('Generating ' +
                        haz_model +
                        ' hazard.')
            hazard = StormEurope.from_icon_grib(
                run_datetime,
                event_date=event_date,
                delete_raw_data=False,
                model_name=full_model_name_temp
                )
            if save_haz:
                hazard.write_hdf5(haz_file_name)
    else:
        raise NotImplementedError("specific 'WS' hazard not implemented yet. " +
                                  "Please specify a valid value for haz_model.")

    # check if hazard is successfully generated for Forecast
    if not isinstance(hazard, Hazard):
        LOGGER.warning('Hazard generation unsuccessful.')
    return hazard, haz_model, run_datetime, event_date



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Functions to download weather forecasts of the ICON weather forecast model
from the German Weather Service DWD (opendata.dwd.de). Currently used for wind
gust forecast in climada.hazard.storm_europe. Other parameters like rainfall or
temperature are available.
"""

__all__ = [
    'download_icon_grib',
    'delete_icon_grib',
    'download_icon_centroids_file',
]


import logging
from pathlib import Path
import bz2
import numpy as np
import datetime as dt

from climada.util.config import CONFIG
from climada.util.files_handler import download_file

LOGGER = logging.getLogger(__name__)



def download_icon_grib(run_datetime,
                       model_name='icon-eu-eps',
                       parameter_name='vmax_10m',
                       max_lead_time=None,
                       download_dir=None):
    """download the gribfiles of a weather forecast run for a certain
    weather parameter from opendata.dwd.de/weather/nwp/.

    Parameters
    ----------
    run_datetime : datetime
        The starting timepoint of the forecast run
    model_name : str
        the name of the forecast model written as it appears
        in the folder structure in opendata.dwd.de/weather/nwp/ or 'test'
    parameter_name : str
        the name of the meteorological parameter
        written as it appears in the folder structure in
        opendata.dwd.de/weather/nwp/
    max_lead_time : int
        number of hours for which files should be
        downloaded, will default to maximum available data
    download_dir: : str or Path
        directory where the downloaded files
        should be saved in

    Returns
    -------
    file_names : list
        a list of filenames that link to all just
        downloaded or available files from the forecast run, defined by
        the input parameters
    """

    LOGGER.info(('Downloading icon grib files of model ' +
                 model_name + ' for parameter ' + parameter_name +
                 ' with starting date ' + run_datetime.strftime('%Y%m%d%H') +
                 '.'))

    url, file_name, lead_times = _create_icon_grib_name(run_datetime,
                                                        model_name,
                                                        parameter_name,
                                                        max_lead_time)
    download_path = CONFIG.local_data.save_dir.dir() if download_dir is None else Path(download_dir)

    #download all files
    file_names = []
    for lead_i in lead_times:
        file_name_i = file_name.format(lead_i=lead_i)
        bz2_pathfile_i = download_path.absolute().joinpath(file_name_i)

        # download file if it does not exist already
        if not bz2_pathfile_i.exists():
            try:
                download_file(url + file_name_i,
                              download_dir=download_dir)
            except Exception as err:
                err_msg = ""
                if run_datetime > (dt.datetime.utcnow()-dt.timedelta(hours=6)):
                    err_msg += (f'Forecast file {file_name_i} might not yet be available '
                                f'on {url}. Wait a few hours. ')
                elif run_datetime < (dt.datetime.utcnow()
                                     -dt.timedelta(hours=24)):
                    err_msg += (f'Forecast file {file_name_i} might no longer be available '
                                 f'on {url}. Files are only openly available for 24 hours. ')
                err_msg += f"Error while downloading {url + file_name_i}: "
                raise type(err)(err_msg + str(err)) from err
        file_names.append(str(bz2_pathfile_i))
    return file_names



def delete_icon_grib(run_datetime,
                     model_name='icon-eu-eps',
                     parameter_name='vmax_10m',
                     max_lead_time=None,
                     download_dir=None):
    """delete the downloaded gribfiles of a weather forecast run for a
    certain weather parameter from opendata.dwd.de/weather/nwp/.

    Parameters
    ----------
    run_datetime : datetime
        The starting timepoint of the forecast run
    model_name : str
        the name of the forecast model written as it appears
        in the folder structure in opendata.dwd.de/weather/nwp/
    parameter_name : str
        the name of the meteorological parameter
        written as it appears in the folder structure in
        opendata.dwd.de/weather/nwp/
    max_lead_time : int
        number of hours for which files should be
        deleted, will default to maximum available data
    download_dir : str or Path
        directory where the downloaded files
        are stored at the moment
    """

    _, file_name, lead_times = _create_icon_grib_name(run_datetime,
                                                      model_name,
                                                      parameter_name,
                                                      max_lead_time)
    download_path = CONFIG.local_data.save_dir.dir() if download_dir is None else Path(download_dir)
    #delete all files
    for lead_i in lead_times:
        file_name_i = file_name.format(lead_i=lead_i)
        full_path_name_i = download_path.absolute().joinpath(file_name_i)
        if full_path_name_i.exists():
            full_path_name_i.unlink()
        else:
            LOGGER.warning('File %s does not exist and could not be deleted.',
                           full_path_name_i)


def _create_icon_grib_name(run_datetime,
                           model_name='icon-eu-eps',
                           parameter_name='vmax_10m',
                           max_lead_time=None):
    """create all parameters to download or delete gribfiles of a weather
    forecast run for a certain weather parameter from
    opendata.dwd.de/weather/nwp/.

    Parameters
    ----------
    run_datetime : datetime
        The starting timepoint of the forecast run
    model_name : str
        the name of the forecast model written as it appears
        in the folder structure in opendata.dwd.de/weather/nwp/
    parameter_name : str
        the name of the meteorological parameter
        written as it appears in the folder structure in
        opendata.dwd.de/weather/nwp/
    max_lead_time : int
        number of hours for which files should be
        selected, will default to maximum available data

    Returns
    -------
    url : str
        url where the gribfiles are stored on opendata.dwd.de
    file_name : str
        filenames of gribfiles (lead_time missing)
    lead_times : np.array
        array of integers representing the leadtimes
        in hours, which are available for download
    """
    # define defaults of the url for each model and parameter combination
    if (model_name == 'icon-eu-eps') & (parameter_name == 'vmax_10m'):
        file_extension = '_europe_icosahedral_single-level_' #this string completes the filename on the server
        file_extension_2 = '' #this string completes the filename on the server
        max_lead_time_default = 120 # maximum available data
        lead_times = np.concatenate((np.arange(1, 49),
                                     np.arange(51, 73, 3),
                                     np.arange(78, 121, 6)
                                     ))
    elif (model_name == 'icon-d2-eps') & (parameter_name == 'vmax_10m'):
        file_extension = '_germany_icosahedral_single-level_' #this string completes the filename on the server
        file_extension_2 = '_2d' #this string completes the filename on the server
        max_lead_time_default = 48 # maximum available data
        lead_times = np.concatenate((np.arange(1, 49),
                                     ))
    elif model_name == 'test':
        file_extension = '_storm_europe_icon_' #this string completes the filename on the server
        file_extension_2 = '' #this string completes the filename on the server
        max_lead_time_default = 2 # maximum available data
        lead_times = np.concatenate((np.arange(1, 49),
                                     np.arange(51, 73, 3),
                                     np.arange(78, 121, 6)
                                     ))
    else:
        raise ValueError(f'Download for model {model_name} and parameter {parameter_name} '
                         'is not yet implemented. '
                         'Please define the default values in the code first.')

    # create the url for download
    url = ('https://opendata.dwd.de/weather/nwp/' +
           model_name +
           '/grib/' +
           run_datetime.strftime('%H') +
           '/' +
           parameter_name +
           '/')
    file_name = (model_name +
                 file_extension +
                 run_datetime.strftime('%Y%m%d%H') +
                 '_' +
                 '{lead_i:03}' +
                 file_extension_2 +
                 '_' +
                 parameter_name +
                 '.grib2.bz2')


    # define the leadtimes
    if  not max_lead_time:
        max_lead_time = max_lead_time_default
    elif max_lead_time > max_lead_time_default:
        LOGGER.warning(('Parameter max_lead_time ' +
                        str(max_lead_time) + ' is bigger than maximum ' +
                        'available files. max_lead_time is adjusted to ' +
                        str(max_lead_time_default)))
        max_lead_time = max_lead_time_default
    lead_times = lead_times[lead_times<=max_lead_time]

    return url, file_name, lead_times


def download_icon_centroids_file(model_name='icon-eu-eps',
                                 download_dir = None):
    """ create centroids based on netcdf files provided by dwd, links
    found here:
    https://www.dwd.de/DE/leistungen/opendata/neuigkeiten/opendata_dez2018_02.html
    https://www.dwd.de/DE/leistungen/opendata/neuigkeiten/opendata_aug2020_01.html

    Parameters
    ----------
    model_name : str
        the name of the forecast model written as it appears
        in the folder structure in opendata.dwd.de/weather/nwp/
    download_dir : str or Path
        directory where the downloaded files
        should be saved in

    Returns
    -------
    file_name : str
        absolute path and filename of the downloaded
        and decompressed netcdf file
    """

    # define url and filename
    url = 'https://opendata.dwd.de/weather/lib/cdo/'
    if model_name == 'icon-eu-eps':
        file_name = 'icon_grid_0028_R02B07_N02.nc.bz2'
    elif model_name == 'icon-eu':
        file_name = 'icon_grid_0024_R02B06_G.nc.bz2'
    elif model_name in ('icon-d2-eps', 'icon-d2'):
        file_name = 'icon_grid_0047_R19B07_L.nc.bz2'
    elif model_name == 'test':
        file_name = 'test_storm_europe_icon_grid.nc.bz2'
    else:
        raise ValueError(f'Creation of centroids for the icon model {model_name} '
                         'is not yet implemented. Please define '
                         'the default values in the code first.')
    download_path = CONFIG.local_data.save_dir.dir() if download_dir is None else Path(download_dir)
    bz2_pathfile = download_path.absolute().joinpath(file_name)
    nc_pathfile = bz2_pathfile.with_suffix('')

    # download and unzip file
    if not nc_pathfile.exists():
        if not bz2_pathfile.exists():
            try:
                download_file(url + file_name,
                              download_dir=download_path)
            except ValueError as err:
                raise ValueError('Error while downloading %s.' % (url + file_name))
        with open(bz2_pathfile, 'rb') as source, open(nc_pathfile, 'wb') as dest:
            dest.write(bz2.decompress(source.read()))
        bz2_pathfile.unlink()

    return str(nc_pathfile)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define DiscRates class.
"""

__all__ = ['DiscRates']

import copy
from array import array
import logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import xlsxwriter

import climada.util.checker as u_check
from climada.entity.tag import Tag
import climada.util.finance as u_fin
import climada.util.hdf5_handler as u_hdf5

LOGGER = logging.getLogger(__name__)

"""MATLAB variable names"""
DEF_VAR_MAT = {'sup_field_name': 'entity',
               'field_name': 'discount',
               'var_name': {'year': 'year',
                            'disc': 'discount_rate'
                           }
              }

"""Excel variable names"""
DEF_VAR_EXCEL = {'sheet_name': 'discount',
                 'col_name': {'year': 'year',
                              'disc': 'discount_rate'
                             }
                }


class DiscRates():
    """
    Defines discount rates and basic methods. Loads from
    files with format defined in FILE_EXT.

    Attributes
    ---------
    tag: Tag
        information about the source data
    years: np.array
        list of years
    rates: np.array
        list of discount rates for each year (between 0 and 1)
    """

    def __init__(self, years=None, rates=None, tag=None):
        """
        Fill discount rates with values and check consistency data

        Parameters
        ----------
        years : numpy.ndarray(int)
            Array of years. Default is numpy.array([]).
        rates : numpy.ndarray(float)
            Discount rates for each year in years.
            Default is numpy.array([]).
            Note: rates given in float, e.g., to set 1% rate use 0.01
        tag : climate.entity.tag
            Metadata. Default is None.
        """
        years = np.array([]) if years is None else years
        rates = np.array([]) if rates is None else rates

        self.years = years
        self.rates = rates
        tag = Tag() if tag is None else tag
        self.tag = tag

    def clear(self):
        """Reinitialize attributes."""

        self.tag = Tag()
        # Following values are given for each defined year
        self.years = np.array([], int)
        self.rates = np.array([], float)

    def check(self):
        """
        Check attributes consistency.

        Raises
        ------
        ValueError
        """
        u_check.size(len(self.years), self.rates, 'DiscRates.rates')

    def select(self, year_range):
        """
        Select discount rates in given years.

        Parameters
        ----------
        year_range: np.array(int)
            continuous sequence of selected years.

        Returns: climada.entity.DiscRates
            The selected discrates in the year_range
        """
        pos_year = np.isin(year_range, self.years)
        if not np.all(pos_year):
            LOGGER.info('No discount rates for given years.')
            return None
        pos_year = np.isin(self.years, year_range)

        return DiscRates(years=self.years[pos_year],
                         rates=self.rates[pos_year],
                         tag=self.tag)

    def append(self, disc_rates):
        """
        Check and append discount rates to current DiscRates. Overwrite
        discount rate if same year.

        Parameters
        ----------
        disc_rates: climada.entity.DiscRates
            DiscRates instance to append

        Raises
        ------
        ValueError
        """
        disc_rates.check()
        if self.years.size == 0:
            self.__dict__ = copy.deepcopy(disc_rates.__dict__)
            return

        self.tag.append(disc_rates.tag)

        new_year = array('l')
        new_rate = array('d')
        for year, rate in zip(disc_rates.years, disc_rates.rates):
            found = np.where(year == self.years)[0]
            if found.size > 0:
                self.rates[found[0]] = rate
            else:
                new_year.append(year)
                new_rate.append(rate)

        self.years = np.append(self.years, new_year).astype(int, copy=False)
        self.rates = np.append(self.rates, new_rate)

    def net_present_value(self, ini_year, end_year, val_years):
        """
        Compute net present value between present year and future year.

        Parameters
        ----------
        ini_year: float
            initial year
        end_year: float
            end year
        val_years: np.array
            cash flow at each year btw ini_year and end_year (both included)

        Returns
        -------
            net_present_value: float
                net present value between present year and future year.

        """
        year_range = np.arange(ini_year, end_year + 1)
        if year_range.size != val_years.size:
            raise ValueError('Wrong size of yearly values.')
        sel_disc = self.select(year_range)
        if sel_disc is None:
            raise ValueError('No information of discount rates for provided years:'
                             f' {ini_year} - {end_year}')
        return u_fin.net_present_value(sel_disc.years, sel_disc.rates,
                                       val_years)

    def plot(self, axis=None, figsize=(6, 8), **kwargs):
        """
        Plot discount rates per year.

        Parameters
        ----------
        axis: matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        figsize: tuple(int, int), optional
            size of the figure. The default is (6,8)
        kwargs: optional
            keyword arguments  passed to plotting function axis.plot

        Returns
        -------
        axis: matplotlib.axes._subplots.AxesSubplot
            axis handles of the plot
        """
        if not axis:
            _, axis = plt.subplots(1, 1, figsize=figsize)

        axis.set_title('Discount rates')
        axis.set_xlabel('Year')
        axis.set_ylabel('discount rate (%)')
        axis.plot(self.years, self.rates * 100, **kwargs)
        axis.set_xlim((self.years.min(), self.years.max()))
        return axis

    @classmethod
    def from_mat(cls, file_name, description='', var_names=DEF_VAR_MAT):
        """
        Read MATLAB file generated with previous MATLAB CLIMADA version.

        Parameters
        ----------
        file_name: str
            filename including path and extension
        description: str, optional
            description of the data. The default is ''
        var_names: dict, optional
            name of the variables in the file. The Default is
            DEF_VAR_MAT = {'sup_field_name': 'entity', 'field_name': 'discount',
               'var_name': {'year': 'year', 'disc': 'discount_rate'}}

        Returns
        -------
        climada.entity.DiscRates :
            The disc rates from matlab
        """

        disc = u_hdf5.read(file_name)
        tag = Tag(file_name=str(file_name), description=description)
        try:
            disc = disc[var_names['sup_field_name']]
        except KeyError:
            pass

        try:
            disc = disc[var_names['field_name']]
            years = np.squeeze(disc[var_names['var_name']['year']]). \
                astype(int, copy=False)
            rates = np.squeeze(disc[var_names['var_name']['disc']])
        except KeyError as err:
            raise KeyError("Not existing variable: %s" % str(err)) from err

        return cls(years=years, rates=rates, tag=tag)

    def read_mat(self, *args, **kwargs):
        """This function is deprecated, use DiscRates.from_mats instead."""
        LOGGER.warning("The use of DiscRates.read_mats is deprecated."
                       "Use DiscRates.from_mats instead.")
        self.__dict__ = DiscRates.from_mat(*args, **kwargs).__dict__

    @classmethod
    def from_excel(cls, file_name, description='', var_names=DEF_VAR_EXCEL):
        """
        Read excel file following template and store variables.

        Parameters
        ----------
        file_name: str
            filename including path and extension
        description: str, optional
            description of the data. The default is ''
        var_names: dict, optional
            name of the variables in the file. The Default is
            DEF_VAR_EXCEL = {'sheet_name': 'discount',
               'col_name': {'year': 'year', 'disc': 'discount_rate'}}

        Returns
        -------
        climada.entity.DiscRates :
            The disc rates from excel
        """

        dfr = pd.read_excel(file_name, var_names['sheet_name'])
        tag = Tag(file_name=str(file_name), description=description)
        try:
            years = dfr[var_names['col_name']['year']].values. \
                astype(int, copy=False)
            rates = dfr[var_names['col_name']['disc']].values
        except KeyError as err:
            raise KeyError("Not existing variable: %s" % str(err)) from err

        return cls(years=years, rates=rates, tag=tag)

    def read_excel(self, *args, **kwargs):
        """This function is deprecated, use DiscRates.from_excel instead."""
        LOGGER.warning("The use of DiscRates.read_excel is deprecated."
                       "Use DiscRates.from_excel instead.")
        self.__dict__ = DiscRates.from_mat(*args, **kwargs).__dict__


    def write_excel(self, file_name, var_names=DEF_VAR_EXCEL):
        """
        Write excel file following template.

        Parameters
        ----------
        file_name: str
            filename including path and extension
        var_names: dict, optional
            name of the variables in the file. The Default is
            DEF_VAR_EXCEL = {'sheet_name': 'discount',
               'col_name': {'year': 'year', 'disc': 'discount_rate'}}
        """

        disc_wb = xlsxwriter.Workbook(file_name)
        disc_ws = disc_wb.add_worksheet(var_names['sheet_name'])

        header = [var_names['col_name']['year'], var_names['col_name']['disc']]
        for icol, head_dat in enumerate(header):
            disc_ws.write(0, icol, head_dat)
        for i_yr, (disc_yr, disc_rt) in enumerate(zip(self.years, self.rates), 1):
            disc_ws.write(i_yr, 0, disc_yr)
            disc_ws.write(i_yr, 1, disc_rt)
        disc_wb.close()



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Data API client
"""
from dataclasses import dataclass
from datetime import datetime
import json
import logging
from pathlib import Path
from urllib.parse import quote, unquote
import time

import pandas as pd
from peewee import CharField, DateTimeField, IntegrityError, Model, SqliteDatabase
import requests
import pycountry

from climada import CONFIG
from climada.entity import Exposures
from climada.hazard import Hazard
from climada.util.constants import SYSTEM_DIR

LOGGER = logging.getLogger(__name__)

DB = SqliteDatabase(Path(CONFIG.data_api.cache_db.str()).expanduser())

HAZ_TYPES = [ht.str() for ht in CONFIG.data_api.supported_hazard_types.list()]
EXP_TYPES = [et.str() for et in CONFIG.data_api.supported_exposures_types.list()]


class Download(Model):
    """Database entry keeping track of downloaded files from the CLIMADA data API"""
    url = CharField()
    path = CharField(unique=True)
    startdownload = DateTimeField()
    enddownload = DateTimeField(null=True)

    class Meta:
        """SQL database and table definition."""
        database = DB

    class Failed(Exception):
        """The download failed for some reason."""


DB.connect()
DB.create_tables([Download])


@dataclass
class FileInfo():
    """file data from CLIMADA data API."""
    uuid:str
    url:str
    file_name:str
    file_format:str
    file_size:int
    check_sum:str


@dataclass
class DataTypeInfo():
    """data type meta data from CLIMADA data API."""
    data_type:str
    data_type_group:str
    status: str
    description:str
    properties:list  # of dict


@dataclass
class DataTypeShortInfo():
    """data type name and group from CLIMADA data API."""
    data_type:str
    data_type_group:str


@dataclass
class DatasetInfo():
    """dataset data from CLIMADA data API."""
    uuid:str
    data_type:DataTypeShortInfo
    name:str
    version:str
    status:str
    properties:dict
    files:list  # of FileInfo
    doi:str
    description:str
    license: str
    activation_date:str
    expiration_date:str

    @staticmethod
    def from_json(jsono):
        """creates a DatasetInfo object from the json object returned by the
        CLIMADA data api server.

        Parameters
        ----------
        jsono : dict

        Returns
        -------
        DatasetInfo
        """
        dataset = DatasetInfo(**jsono)
        dataset.data_type = DataTypeShortInfo(data_type=dataset.data_type['data_type'],
                                              data_type_group=dataset.data_type['data_type_group'])
        dataset.files = [FileInfo(uuid=dataset.uuid, **filo) for filo in dataset.files]
        return dataset


def checksize(local_path, fileinfo):
    """Checks sanity of downloaded file simply by comparing actual and registered size.

    Parameters
    ----------
    local_path : Path
        the downloaded file
    filinfo : FileInfo
        file information from CLIMADA data API

    Raises
    ------
    Download.Failed
        if the file is not what it's supposed to be
    """
    if not local_path.is_file():
        raise Download.Failed(f"{str(local_path)} is not a file")
    if local_path.stat().st_size != fileinfo.file_size:
        raise Download.Failed(f"{str(local_path)} has the wrong size:"
            f"{local_path.stat().st_size} instead of {fileinfo.file_size}")


def checkhash(local_path, fileinfo):
    """Checks sanity of downloaded file by comparing actual and registered check sum.

    Parameters
    ----------
    local_path : Path
        the downloaded file
    filinfo : FileInfo
        file information from CLIMADA data API

    Raises
    ------
    Download.Failed
        if the file is not what it's supposed to be
    """
    raise NotImplementedError("sanity check by hash sum needs to be implemented yet")


class Client():
    """Python wrapper around REST calls to the CLIMADA data API server.
    """
    MAX_WAITING_PERIOD = 6
    UNLIMITED = 100000

    class AmbiguousResult(Exception):
        """Custom Exception for Non-Unique Query Result"""

    class NoResult(Exception):
        """Custom Exception for No Query Result"""

    def __init__(self):
        """Constructor of Client.

        Data API host and chunk_size (for download) are configurable values.
        Default values are 'climada.ethz.ch' and 8096 respectively.
        """
        self.headers = {"accept": "application/json"}
        self.url = CONFIG.data_api.url.str().rstrip("/")
        self.chunk_size = CONFIG.data_api.chunk_size.int()

    @staticmethod
    def _request_200(url, **kwargs):
        """Helper method, triaging successfull and failing requests.

        Returns
        -------
        dict
            loaded from the json object of a successful request.

        Raises
        ------
        NoResult
            if the response status code is different from 200
        """
        page = requests.get(url, **kwargs)
        if page.status_code == 200:
            return json.loads(page.content.decode())
        raise Client.NoResult(page.content.decode())

    @staticmethod
    def _divide_straight_from_multi(properties):
        straights, multis = dict(), dict()
        for k, _v in properties.items():
            if isinstance(_v, str):
                straights[k] = _v
            elif isinstance(_v, list):
                multis[k] = _v
            else:
                raise ValueError("properties must be a string or a list of strings")
        return straights, multis

    @staticmethod
    def _filter_datasets(datasets, multi_props):
        pdf = pd.DataFrame([ds.properties for ds in datasets])
        for prop, selection in multi_props.items():
            pdf = pdf[pdf[prop].isin(selection)]
        return [datasets[i] for i in pdf.index]

    def list_dataset_infos(self, data_type=None, name=None, version=None, properties=None,
                     status='active'):
        """Find all datasets matching the given parameters.

        Parameters
        ----------
        data_type : str, optional
            data_type of the dataset, e.g., 'litpop' or 'draught'
        name : str, optional
            the name of the dataset
        version : str, optional
            the version of the dataset
        properties : dict, optional
            search parameters for dataset properties, by default None
            any property has a string for key and can be a string or a list of strings for value
        status : str, optional
            valid values are 'preliminary', 'active', 'expired', 'test_dataset' and None
            by default 'active'

        Returns
        -------
        list of DatasetInfo
        """
        url = f'{self.url}/dataset'
        params = {
            'data_type': data_type,
            'name': name,
            'version': version,
            'status': '' if status is None else status,
            'limit': Client.UNLIMITED,
        }

        if properties:
            straight_props, multi_props = self._divide_straight_from_multi(properties)
        else:
            straight_props, multi_props = None, None

        if straight_props:
            params.update(straight_props)

        datasets = [DatasetInfo.from_json(ds) for ds in Client._request_200(url, params=params)]

        if datasets and multi_props:
            return self._filter_datasets(datasets, multi_props)
        return datasets

    def get_dataset_info(self, data_type=None, name=None, version=None, properties=None,
                    status='active'):
        """Find the one dataset that matches the given parameters.

        Parameters
        ----------
        data_type : str, optional
            data_type of the dataset, e.g., 'litpop' or 'draught'
        name : str, optional
            the name of the dataset
        version : str, optional
            the version of the dataset
        properties : dict, optional
            search parameters for dataset properties, by default None
            any property has a string for key and can be a string or a list of strings for value
        status : str, optional
            valid values are 'preliminary', 'active', 'expired', 'test_dataset', None
            by default 'active'

        Returns
        -------
        DatasetInfo

        Raises
        ------
        AmbiguousResult
            when there is more than one dataset matching the search parameters
        NoResult
            when there is no dataset matching the search parameters
        """
        jarr = self.list_dataset_infos(data_type=data_type, name=name, version=version,
                                 properties=properties, status=status)
        if len(jarr) > 1:
            raise Client.AmbiguousResult("there are several datasets meeting the requirements:"
                                        f" {jarr}")
        if len(jarr) < 1:
            raise Client.NoResult("there is no dataset meeting the requirements")
        return jarr[0]

    def get_dataset_info_by_uuid(self, uuid):
        """Returns the data from 'https://climada.ethz.ch/data-api/v1/dataset/{uuid}' as DatasetInfo object.

        Parameters
        ----------
        uuid : str
            the universal unique identifier of the dataset

        Returns
        -------
        DatasetInfo

        Raises
        ------
        NoResult
            if the uuid is not valid
        """
        url = f'{self.url}/dataset/{uuid}'
        return DatasetInfo.from_json(Client._request_200(url))

    def list_data_type_infos(self, data_type_group=None):
        """Returns all data types from the climada data API
        belonging to a given data type group.

        Parameters
        ----------
        data_type_group : str, optional
            name of the data type group, by default None

        Returns
        -------
        list of DataTypeInfo
        """
        url = f'{self.url}/data_type'
        params = {'data_type_group': data_type_group} \
            if data_type_group else {}
        return [DataTypeInfo(**jobj) for jobj in Client._request_200(url, params=params)]

    def get_data_type_info(self, data_type):
        """Returns the metadata of the data type with the given name from the climada data API.

        Parameters
        ----------
        data_type : str
            data type name

        Returns
        -------
        DataTypeInfo

        Raises
        ------
        NoResult
            if there is no such data type registered
        """
        url = f'{self.url}/data_type/{quote(data_type)}'
        return DataTypeInfo(**Client._request_200(url))

    def _download(self, url, path, replace=False):
        """Downloads a file from the given url to a specified location.

        Parameters
        ----------
        url : str
            the link to the file to be downloaded
        path : Path
            download path, if it's a directory the original file name is kept
        replace : bool, optional
            flag to indicate whether a present file with the same name should
            be replaced

        Returns
        -------
        Path
            Path to the downloaded file

        Raises
        ------
        FileExistsError
            in case there is already a file present at the given location
            and replace is False
        """
        if path.is_dir():
            path /= unquote(url.split('/')[-1])
        if path.is_file() and not replace:
            raise FileExistsError(path)
        with requests.get(url, stream=True) as stream:
            stream.raise_for_status()
            with open(path, 'wb') as dump:
                for chunk in stream.iter_content(chunk_size=self.chunk_size):
                    dump.write(chunk)
        return path

    def _tracked_download(self, remote_url, local_path):
        if local_path.is_dir():
            raise Exception("tracked download requires a path to a file not a directory")
        path_as_str = str(local_path.absolute())
        try:
            dlf = Download.create(url=remote_url, path=path_as_str, startdownload=datetime.utcnow())
        except IntegrityError as ierr:
            dlf = Download.get(Download.path==path_as_str)
            if dlf.url != remote_url:
                raise Exception("this file has been downloaded from another url, "
                    "please purge the entry from data base before trying again") from ierr
            return dlf
        try:
            self._download(url=remote_url, path=local_path, replace=True)
            dlf.enddownload = datetime.utcnow()
            dlf.save()
        except Exception:
            dlf.delete_instance()
            raise
        return Download.get(Download.path==path_as_str)

    def _download_file(self, local_path, fileinfo, check=checksize, retries=3):
        """Download a file if it is not already present at the target destination.

        Parameters
        ----------
        local_path : Path
            target destination,
            if it is a directory the original filename (fileinfo.filen_name) is kept
        fileinfo : FileInfo
            file object as retrieved from the data api
        check : function, optional
            how to check download success, by default checksize
        retries : int, optional
            how many times one should retry in case of failure, by default 3

        Returns
        -------
        Path
            the path to the downloaded file

        Raises
        ------
        Exception
            when number of retries was exceeded or when a download is already running
        """
        try:
            if local_path.is_dir():
                local_path /= fileinfo.file_name
            downloaded = self._tracked_download(remote_url=fileinfo.url, local_path=local_path)
            if not downloaded.enddownload:
                raise Download.Failed("Download seems to be in progress, please try again later"
                    f" or remove cache entry by calling `purge_cache(Path('{local_path}'))`!")
            try:
                check(local_path, fileinfo)
            except Download.Failed as dlf:
                local_path.unlink(missing_ok=True)
                self.purge_cache(local_path)
                raise dlf
            return local_path
        except Download.Failed as dle:
            if retries < 1:
                raise dle
            LOGGER.warning("Download failed: %s, retrying...", dle)
            time.sleep(self.MAX_WAITING_PERIOD/retries)
            return self._download_file(local_path=local_path, fileinfo=fileinfo, check=check,
                                       retries=retries - 1)

    def download_dataset(self, dataset, target_dir=SYSTEM_DIR, organize_path=True):
        """Download all files from a given dataset to a given directory.

        Parameters
        ----------
        dataset : DatasetInfo
            the dataset
        target_dir : Path, optional
            target directory for download, by default `climada.util.constants.SYSTEM_DIR`
        organize_path: bool, optional
            if set to True the files will end up in subdirectories of target_dir:
            [target_dir]/[data_type_group]/[data_type]/[name]/[version]
            by default True

        Returns
        -------
        download_dir : Path
            the path to the directory containing the downloaded files,
            will be created if organize_path is True
        downloaded_files : list of Path
            the downloaded files themselves

        Raises
        ------
        Exception
            when one of the files cannot be downloaded
        """
        if not target_dir.is_dir():
            raise ValueError(f"{target_dir} is not a directory")

        if organize_path:
            target_dir = self._organize_path(dataset, target_dir)

        return target_dir, [
            self._download_file(local_path=target_dir, fileinfo=dsfile)
            for dsfile in dataset.files
        ]

    @staticmethod
    def _organize_path(dataset, target_dir):
        if dataset.data_type.data_type_group:
            target_dir /= dataset.data_type.data_type_group
        if dataset.data_type.data_type_group != dataset.data_type.data_type:
            target_dir /= dataset.data_type.data_type
        target_dir /= dataset.name
        if dataset.version:
            target_dir /= dataset.version
        target_dir.mkdir(exist_ok=True, parents=True)
        return target_dir

    @staticmethod
    def purge_cache(local_path):
        """Removes entry from the sqlite database that keeps track of files downloaded by
        `cached_download`. This may be necessary in case a previous attempt has failed
        in an uncontroled way (power outage or the like).

        Parameters
        ----------
        local_path : Path
            target destination
        fileinfo : FileInfo
            file object as retrieved from the data api
        """
        dlf = Download.get(Download.path==str(local_path.absolute()))
        dlf.delete_instance()

    @staticmethod
    def _multi_version(datasets):
        ddf = pd.DataFrame(datasets)
        gdf = ddf.groupby('name').agg({'version': 'nunique'})
        return list(gdf[gdf.version > 1].index)

    def get_hazard(self, hazard_type, name=None, version=None, properties=None,
                   status='active', dump_dir=SYSTEM_DIR):
        """Queries the data api for hazard datasets of the given type, downloads associated
        hdf5 files and turns them into a climada.hazard.Hazard object.

        Parameters
        ----------
        hazard_type : str
            Type of climada hazard.
        name : str, optional
            the name of the dataset
        version : str, optional
            the version of the dataset
        properties : dict, optional
            search parameters for dataset properties, by default None
            any property has a string for key and can be a string or a list of strings for value
        status : str, optional
            valid values are 'preliminary', 'active', 'expired', 'test_dataset', None
            by default 'active'
        dump_dir : str, optional
            Directory where the files should be downoladed. Default: SYSTEM_DIR
            If the directory is the SYSTEM_DIR (as configured in
            climada.conf, i.g. ~/climada/data), the eventual target directory is organized into
            dump_dir > hazard_type > dataset name > version

        Returns
        -------
        climada.hazard.Hazard
            The combined hazard object
        """
        if not hazard_type in HAZ_TYPES:
            raise ValueError("Valid hazard types are a subset of CLIMADA hazard types."
                             f" Currently these types are supported: {HAZ_TYPES}")
        dataset = self.get_dataset_info(data_type=hazard_type, name=name, version=version,
                                        properties=properties, status=status)
        return self.to_hazard(dataset, dump_dir)

    def to_hazard(self, dataset, dump_dir=SYSTEM_DIR):
        """Downloads hdf5 files belonging to the given datasets reads them into Hazards and
        concatenates them into a single climada.Hazard object.

        Parameters
        ----------
        dataset : DatasetInfo
            Dataset to download and read into climada.Hazard object.
        dump_dir : str, optional
            Directory where the files should be downoladed. Default: SYSTEM_DIR (as configured in
            climada.conf, i.g. ~/climada/data).
            If the directory is the SYSTEM_DIR, the eventual target directory is organized into
            dump_dir > hazard_type > dataset name > version

        Returns
        -------
        climada.hazard.Hazard
            The combined hazard object
        """
        target_dir = self._organize_path(dataset, dump_dir) \
                     if dump_dir == SYSTEM_DIR else dump_dir
        
        hazard_list = [
            Hazard.from_hdf5(self._download_file(target_dir, dsf))
            for dsf in dataset.files
            if dsf.file_format == 'hdf5'
        ]
        if not hazard_list:
            raise ValueError("no hdf5 files found in dataset")
        if len(hazard_list) == 1:
            return hazard_list[0]
        hazard_concat = Hazard()
        hazard_concat = hazard_concat.concat(hazard_list)
        hazard_concat.sanitize_event_ids()
        hazard_concat.check()
        return hazard_concat

    def get_exposures(self, exposures_type, name=None, version=None, properties=None,
                   status='active', dump_dir=SYSTEM_DIR):
        """Queries the data api for exposures datasets of the given type, downloads associated
        hdf5 files and turns them into a climada.entity.exposures.Exposures object.

        Parameters
        ----------
        exposures_type : str
            Type of climada exposures.
        name : str, optional
            the name of the dataset
        version : str, optional
            the version of the dataset
        properties : dict, optional
            search parameters for dataset properties, by default None
            any property has a string for key and can be a string or a list of strings for value
        status : str, optional
            valid values are 'preliminary', 'active', 'expired', 'test_dataset', None
            by default 'active'
        dump_dir : str, optional
            Directory where the files should be downoladed. Default: SYSTEM_DIR
            If the directory is the SYSTEM_DIR, the eventual target directory is organized into
            dump_dir > hazard_type > dataset name > version

        Returns
        -------
        climada.entity.exposures.Exposures
            The combined exposures object
        """
        if not exposures_type in EXP_TYPES:
            raise ValueError("Valid exposures types are a subset of CLIMADA exposures types."
                             f" Currently these types are supported: {EXP_TYPES}")
        dataset = self.get_dataset_info(data_type=exposures_type, name=name, version=version,
                                        properties=properties, status=status)
        return self.to_exposures(dataset, dump_dir)

    def to_exposures(self, dataset, dump_dir=SYSTEM_DIR):
        """Downloads hdf5 files belonging to the given datasets reads them into Exposures and
        concatenates them into a single climada.Exposures object.

        Parameters
        ----------
        dataset : DatasetInfo
            Dataset to download and read into climada.Exposures objects.
        dump_dir : str, optional
            Directory where the files should be downoladed. Default: SYSTEM_DIR (as configured in
            climada.conf, i.g. ~/climada/data).
            If the directory is the SYSTEM_DIR, the eventual target directory is organized into
            dump_dir > exposures_type > dataset name > version

        Returns
        -------
        climada.entity.exposures.Exposures
            The combined exposures object
        """
        target_dir = self._organize_path(dataset, dump_dir) \
                     if dump_dir == SYSTEM_DIR else dump_dir
        exposures_list = [
            Exposures.from_hdf5(self._download_file(target_dir, dsf))
            for dsf in dataset.files
            if dsf.file_format == 'hdf5'
        ]
        if not exposures_list:
            raise ValueError("no hdf5 files found in dataset")
        if len(exposures_list) == 1:
            return exposures_list[0]
        exposures_concat = Exposures()
        exposures_concat = exposures_concat.concat(exposures_list)
        exposures_concat.check()
        return exposures_concat

    def get_litpop_default(self, country=None, dump_dir=SYSTEM_DIR):
        """Get a LitPop instance on a 150arcsec grid with the default parameters:
        exponents = (1,1) and fin_mode = 'pc'.

        Parameters
        ----------
        country : str or list, optional
            List of country name or iso3 codes for which to create the LitPop object.
            If None is given, a global LitPop instance is created. Defaut is None
        dump_dir : str
            directory where the files should be downoladed. Default: SYSTEM_DIR

        Returns
        -------
        climada.entity.exposures.Exposures
            default litpop Exposures object
        """
        properties = {
            'exponents': '(1,1)',
            'fin_mode': 'pc'
        }
        if country is None:
            properties['spatial_coverage'] = 'global'
        elif isinstance(country, str):
            properties['country_name'] = pycountry.countries.lookup(country).name
        elif isinstance(country, list):
            properties['country_name'] = [pycountry.countries.lookup(c).name for c in country]
        else:
            raise ValueError("country must be string or list of strings")
        return self.get_exposures(exposures_type='litpop', dump_dir=dump_dir, properties=properties)

    @staticmethod
    def get_property_values(dataset_infos, known_property_values=None,
                            exclude_properties=None):
        """Returns a dictionnary of possible values for properties of a data type, optionally given known property values.

        Parameters
        ----------
        dataset_infos : list of DataSetInfo
            as returned by list_dataset_infos
        known_properties_value : dict, optional
            dict {'property':'value1, 'property2':'value2'}, to provide only a subset of property
            values that can be combined with the given properties.
        exclude_properties: list of str, optional
            properties in this list will be excluded from the resulting dictionary, e.g., because
            they are strictly metadata and don't provide any information essential to the dataset.
            Default: 'creation_date', 'climada_version'

        Returns
        -------
        dict
            of possibles property values
        """
        if exclude_properties is None:
            exclude_properties = ['date_creation', 'climada_version']

        ppdf = pd.DataFrame([ds.properties for ds in dataset_infos])
        if known_property_values:
            for key, val in known_property_values.items():
                ppdf = ppdf[ppdf[key] == val]
        if not len(ppdf):
            raise Client.NoResult("there is no dataset meeting the requirements")

        property_values = dict()
        for col in ppdf.columns:
            if col in exclude_properties:
                continue
            valar = ppdf[col].dropna().drop_duplicates().values
            if valar.size:
                property_values[col] = list(valar)
        return property_values

    @staticmethod
    def into_datasets_df(dataset_infos):
        """Convenience function providing a DataFrame of datasets with properties.

        Parameters
        ----------
        dataset_infos : list of DatasetInfo
             as returned by list_dataset_infos

        Returns
        -------
        pandas.DataFrame
            of datasets with properties as found in query by arguments
        """
        dsdf = pd.DataFrame(dataset_infos)
        ppdf = pd.DataFrame([ds.properties for ds in dataset_infos])
        dtdf = pd.DataFrame([pd.Series(dt) for dt in dsdf.data_type])

        return dtdf.loc[:, [c for c in dtdf.columns if c not in ['description', 'properties']]].join(
               dsdf.loc[:, [c for c in dsdf.columns if c not in ['data_type', 'properties', 'files']]]).join(
               ppdf)

    @staticmethod
    def into_files_df(dataset_infos):
        """Convenience function providing a DataFrame of files aligned with the input datasets.

        Parameters
        ----------
        datasets : list of DatasetInfo
            as returned by list_dataset_infos

        Returns
        -------
        pandas.DataFrame
            of the files' informations including dataset informations
        """
        return Client.into_datasets_df(dataset_infos) \
            .merge(pd.DataFrame([dsfile for ds in dataset_infos for dsfile in ds.files]))



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Uncertainty Impact class
"""

__all__ = ['CalcImpact']

import logging
import time

import pandas as pd
import numpy as np

from climada.engine import Impact
from climada.engine.unsequa import Calc, InputVar, UncImpactOutput
from climada.util import log_level

LOGGER = logging.getLogger(__name__)


class CalcImpact(Calc):
    """
    Impact uncertainty caclulation class.

    This is the class to perform uncertainty analysis on the outputs of a
    climada.engine.impact.Impact() object.

    Attributes
    ----------
    rp : list(int)
        List of the chosen return periods.
    calc_eai_exp : bool
        Compute eai_exp or not
    calc_at_event : bool
        Compute eai_exp or not
    metric_names : tuple(str)
        Names of the impact output metris
        ('aai_agg', 'freq_curve', 'at_event', 'eai_exp', 'tot_value')
    value_unit : str
        Unit of the exposures value
    input_var_names : tuple(str)
        Names of the required uncertainty input variables
        ('exp_input_var', 'impf_input_var', 'haz_input_var')
    exp_input_var : climada.engine.uncertainty.input_var.InputVar
        Exposure uncertainty variable
    impf_input_var : climada.engine.uncertainty.input_var.InputVar
        Impact function set uncertainty variable
    haz_input_var: climada.engine.uncertainty.input_var.InputVar
        Hazard uncertainty variable
    """

    def __init__(self, exp_input_var, impf_input_var, haz_input_var):
        """Initialize UncCalcImpact

        Sets the uncertainty input variables, the impact metric_names, and the
        units.

        Parameters
        ----------
        exp_input_var : climada.engine.uncertainty.input_var.InputVar or climada.entity.Exposure
            Exposure uncertainty variable or Exposure
        impf_input_var : climada.engine.uncertainty.input_var.InputVar or climada.entity.ImpactFuncSet
            Impact function set uncertainty variable or Impact function set
        haz_input_var : climada.engine.uncertainty.input_var.InputVar or climada.hazard.Hazard
            Hazard uncertainty variable or Hazard

        """

        Calc.__init__(self)
        self.input_var_names = ('exp_input_var', 'impf_input_var', 'haz_input_var')
        self.exp_input_var =  InputVar.var_to_inputvar(exp_input_var)
        self.impf_input_var =  InputVar.var_to_inputvar(impf_input_var)
        self.haz_input_var =  InputVar.var_to_inputvar(haz_input_var)
        self.metric_names = ('aai_agg', 'freq_curve', 'at_event',
                             'eai_exp', 'tot_value')
        self.value_unit = self.exp_input_var.evaluate().value_unit
        self.check_distr()


    def uncertainty(self,
                    unc_sample,
                    rp=None,
                    calc_eai_exp=False,
                    calc_at_event=False,
                    pool=None
                    ):
        """
        Computes the impact for each sample in unc_data.sample_df.

        By default, the aggregated average annual impact
        (impact.aai_agg) and the excees impact at return periods rp
        (imppact.calc_freq_curve(self.rp).impact) is computed.
        Optionally, eai_exp and at_event is computed (this may require
        a larger amount of memory if the number of samples and/or the number
        of centroids and/or exposures points is large).

        This sets the attributes self.rp, self.calc_eai_exp,
        self.calc_at_event, self.metrics.

        This sets the attributes:
        unc_output.aai_agg_unc_df,
        unc_output.freq_curve_unc_df
        unc_output.eai_exp_unc_df
        unc_output.at_event_unc_df
        unc_output.tot_value_unc_df
        unc_output.unit

        Parameters
        ----------
        unc_sample : climada.engine.uncertainty.unc_output.UncOutput
            Uncertainty data object with the input parameters samples
        rp : list(int), optional
            Return periods in years to be computed.
            The default is [5, 10, 20, 50, 100, 250].
        calc_eai_exp : boolean, optional
            Toggle computation of the impact at each centroid location.
            The default is False.
        calc_at_event : boolean, optional
            Toggle computation of the impact for each event.
            The default is False.
        pool : pathos.pools.ProcessPool, optional
            Pool of CPUs for parralel computations.
            The default is None.

        Returns
        -------
        unc_output : climada.engine.uncertainty.unc_output.UncImpactOutput
            Uncertainty data object with the impact outputs for each sample
            and all the sample data copied over from unc_sample.

        Raises
        ------
        ValueError:
            If no sampling parameters defined, the distribution cannot
            be computed.

        See Also
        --------
        climada.engine.impact: Compute risk.

        """

        if unc_sample.samples_df.empty:
            raise ValueError("No sample was found. Please create one first"
                             "using UncImpact.make_sample(N)")

        samples_df = unc_sample.samples_df.copy(deep=True)

        unit = self.value_unit

        if rp is None:
            rp=[5, 10, 20, 50, 100, 250]

        self.rp = rp
        self.calc_eai_exp = calc_eai_exp
        self.calc_at_event = calc_at_event

        start = time.time()
        one_sample = samples_df.iloc[0:1].iterrows()
        imp_metrics = map(self._map_impact_calc, one_sample)
        [aai_agg_list, freq_curve_list,
         eai_exp_list, at_event_list, tot_value_list] = list(zip(*imp_metrics))
        elapsed_time = (time.time() - start)
        self.est_comp_time(unc_sample.n_samples, elapsed_time, pool)

        #Compute impact distributions
        with log_level(level='ERROR', name_prefix='climada'):
            if pool:
                LOGGER.info('Using %s CPUs.', pool.ncpus)
                chunksize = min(unc_sample.n_samples // pool.ncpus, 100)
                imp_metrics = pool.map(self._map_impact_calc,
                                        samples_df.iterrows(),
                                        chunsize = chunksize)

            else:
                imp_metrics = map(self._map_impact_calc,
                                  samples_df.iterrows())

        #Perform the actual computation
        with log_level(level='ERROR', name_prefix='climada'):
            [aai_agg_list, freq_curve_list,
             eai_exp_list, at_event_list,
             tot_value_list] = list(zip(*imp_metrics))

        # Assign computed impact distribution data to self
        aai_agg_unc_df  = pd.DataFrame(aai_agg_list,
                                                columns = ['aai_agg'])
        freq_curve_unc_df = pd.DataFrame(freq_curve_list,
                                    columns=['rp' + str(n) for n in rp])
        eai_exp_unc_df =  pd.DataFrame(eai_exp_list)
        # Setting to sparse dataframes is not compatible with .to_hdf5
        # if np.count_nonzero(df_eai_exp.to_numpy()) / df_eai_exp.size < 0.5:
        #     df_eai_exp = df_eai_exp.astype(pd.SparseDtype("float", 0.0))
        #eai_exp_unc_df = df_eai_exp
        at_event_unc_df = pd.DataFrame(at_event_list)
        # Setting to sparse dataframes is not compatible with .to_hdf5
        # if np.count_nonzero(df_at_event.to_numpy()) / df_at_event.size < 0.5:
        #     df_at_event = df_at_event.astype(pd.SparseDtype("float", 0.0))
        #at_event_unc_df = df_at_event
        tot_value_unc_df = pd.DataFrame(tot_value_list,
                                                 columns = ['tot_value'])

        if calc_eai_exp:
            exp = self.exp_input_var.evaluate()
            coord_df = exp.gdf[['latitude', 'longitude']]
        else:
            coord_df = pd.DataFrame([])

        return UncImpactOutput(samples_df=samples_df,
                               unit=unit,
                               aai_agg_unc_df=aai_agg_unc_df,
                               freq_curve_unc_df=freq_curve_unc_df,
                               eai_exp_unc_df=eai_exp_unc_df,
                               at_event_unc_df=at_event_unc_df,
                               tot_value_unc_df=tot_value_unc_df,
                               coord_df=coord_df
                               )


    def _map_impact_calc(self, sample_iterrows):
        """
        Map to compute impact for all parameter samples in parrallel

        Parameters
        ----------
        sample_iterrows : pd.DataFrame.iterrows()
            Generator of the parameter samples

        Returns
        -------
         : list
            impact metrics list for all samples containing aai_agg, rp_curve,
            eai_exp (np.array([]) if self.calc_eai_exp=False) and at_event
            (np.array([]) if self.calc_at_event=False).

        """

        # [1] only the rows of the dataframe passed by pd.DataFrame.iterrows()
        exp_samples = sample_iterrows[1][self.exp_input_var.labels].to_dict()
        impf_samples = sample_iterrows[1][self.impf_input_var.labels].to_dict()
        haz_samples = sample_iterrows[1][self.haz_input_var.labels].to_dict()

        exp = self.exp_input_var.evaluate(**exp_samples)
        impf = self.impf_input_var.evaluate(**impf_samples)
        haz = self.haz_input_var.evaluate(**haz_samples)

        imp = Impact()

        imp.calc(exposures=exp, impact_funcs=impf, hazard=haz)

        # Extract from climada.impact the chosen metrics
        freq_curve = imp.calc_freq_curve(self.rp).impact

        if self.calc_eai_exp:
            eai_exp = imp.eai_exp
        else:
            eai_exp = np.array([])

        if self.calc_at_event:
            at_event= imp.at_event
        else:
            at_event = np.array([])

        return [imp.aai_agg, freq_curve, eai_exp, at_event, imp.tot_value]



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Forecast
"""

__all__ = ["Forecast"]


import logging
import datetime as dt
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
from matplotlib.ticker import PercentFormatter, ScalarFormatter
from matplotlib.colors import ListedColormap, BoundaryNorm
import cartopy.crs as ccrs
import pyproj
import shapely
from cartopy.io import shapereader
from mpl_toolkits.axes_grid1 import make_axes_locatable

from climada.engine import Impact
import climada.util.plot as u_plot
from climada.util.config import CONFIG
from climada.util.files_handler import to_list
import climada.util.coordinates as u_coord
from climada.util.value_representation import (
    value_to_monetary_unit as u_value_to_monetary_unit,
)

LOGGER = logging.getLogger(__name__)

DATA_DIR = CONFIG.local_data.save_dir.str()

FORECAST_DIR = CONFIG.engine.forecast.local_data.str()

FORECAST_PLOT_DIR = CONFIG.engine.forecast.plot_dir.dir()

# defining colormaps
# The colors are in line the european meteoalarm colors http://www.meteoalarm.info/
# and with the colors used at MeteoSwiss
# https://www.natural-hazards.ch/home/dealing-with-natural-hazards/explanation-of-the-danger-levels.html
# colors for warning levels
COLORS_WARN = np.array(
    [
        [204 / 255, 255 / 255, 102 / 255, 1],  # green
        [255 / 255, 255 / 255, 0 / 255, 1],  # yellow
        [255 / 255, 153 / 255, 0 / 255, 1],  # orange
        [255 / 255, 0 / 255, 0 / 255, 1],  # red
        [128 / 255, 0 / 255, 0 / 255, 1],  # dark red
    ]
)
# colors for warning probabilities
warnprob_colors = np.array(
    [
        [255 / 255, 255 / 255, 255 / 255, 1],  # white
        [215 / 255, 227 / 255, 238 / 255, 1],  # lightest blue
        [181 / 255, 202 / 255, 255 / 255, 1],  # ligthblue
        [127 / 255, 151 / 255, 255 / 255, 1],  # blue
        [171 / 255, 207 / 255, 99 / 255, 1],  # green
        [232 / 255, 245 / 255, 158 / 255, 1],  # light yellow?
        [255 / 255, 250 / 255, 20 / 255, 1],  # yellow
        [255 / 255, 209 / 255, 33 / 255, 1],  # light orange
        [255 / 255, 163 / 255, 10 / 255, 1],  # orange
        [255 / 255, 76 / 255, 0 / 255, 1],  # red
    ]
)
warnprob_colors_extended = np.repeat(warnprob_colors, 10, axis=0)
CMAP_WARNPROB = ListedColormap(warnprob_colors_extended)
# colors for impact forecast
color_map_pre = plt.get_cmap("plasma", 90)
impact_colors = color_map_pre(np.linspace(0, 1, 90))
white_extended = np.repeat([[255 / 255, 255 / 255, 255 / 255, 1]], 10, axis=0)
impact_colors_extended = np.append(white_extended, impact_colors, axis=0)
CMAP_IMPACT = ListedColormap(impact_colors_extended)


class Forecast:
    """Forecast definition. Compute an impact forecast with predefined hazard
    originating from a forecast (like numerical weather prediction models),
    exposure and impact. Use the calc() method to calculate a forecasted
    impact. Then use the plotting methods to illustrate the forecasted impacts.
    By default plots are saved under in a '/forecast/plots' folder in the
    configurable save_dir in local_data (see climada.util.config) under a name
    summarizing the Hazard type, haz model name, initialization time of the
    forecast run, event date, exposure name and the plot title.
    As the class is relatively new, there might be future changes to the attributes,
    the methods, and the parameters used to call the methods.
    It was discovered at some point, that there might be a memory leak in
    matplotlib even when figures are closed
    (https://github.com/matplotlib/matplotlib/issues/8519). Due to this reason
    the plotting functions in this module have the flag close_fig, to close
    figures within the function scope, which might mitigate that problem if a
    script runs this plotting functions many times.

    Attributes
    ----------
    run_datetime : list of datetime.datetime
        initialization time of the forecast model run used to create the Hazard
    event_date: datetime.datetime
        Date on which the Hazard event takes place
    hazard : list of CLIMADA Hazard
        List of the hazard forecast with different lead times.
    haz_model : str
        Short string specifying the model used to create the hazard,
        if possible three big letters.
    exposure : Exposure
        an CLIMADA Exposures containg values at risk
    exposure_name : str
        string specifying the exposure (e.g. 'EU'), which is used to
        name output files.
    vulnerability : ImpactFuncSet
        Set of impact functions used in the impact calculation.
    """

    def __init__(
        self, hazard_dict, exposure, impact_funcs, haz_model="NWP", exposure_name=None
    ):
        """Initialization with hazard, exposure and vulnerability.

        Parameters
        ----------
        hazard_dict : dict
            Dictionary of the format {run_datetime: Hazard} with run_datetime
            being the initialization time of a weather forecast run and Hazard
            being a CLIMADA Hazard derived from that forecast for one event.
            A probabilistic representation of that one event is possible,
            as long as the attribute Hazard.date is the same for all
            events. Several run_datetime:Hazard combinations for the same
            event can be provided.
        exposure : Exposure
        impact_funcs : ImpactFuncSet
        haz_model : str, optional
            Short string specifying the model used to create the hazard,
            if possible three big letters. Default is 'NWP' for numerical
            weather prediction.
        exposure_name : str, optional
            string specifying the exposure (e.g. 'EU'), which is used to
            name output files.
        """
        self.run_datetime = list(hazard_dict.keys())
        self.hazard = list(hazard_dict.values())
        # check event_date
        hazard_date = np.unique(
            [date for hazard in self.hazard for date in hazard.date]
        )
        if not len(hazard_date) == 1:
            raise ValueError(
                "Please provide hazards containing only one "
                + "event_date. The current hazards contain several "
                + "events with different event_dates and the Forecast "
                + "class cannot function proberly with such hazards."
            )
        self.event_date = dt.datetime.fromordinal(hazard_date[0])
        self.haz_model = haz_model
        self.exposure = exposure
        if exposure_name is None:
            try:
                self.exposure_name = u_coord.country_to_iso(
                    exposure.gdf.region_id.unique()[0], "name"
                )
            except (KeyError, AttributeError):
                self.exposure_name = "custom"
        else:
            self.exposure_name = exposure_name
        self.vulnerability = impact_funcs
        self._impact = [Impact() for dt in self.run_datetime]

    def ei_exp(self, run_datetime=None):
        """
        Expected impact per exposure

        Parameters
        ----------
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.
        Returns
        -------
        float
        """
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        return self._impact[haz_ind].eai_exp

    def ai_agg(self, run_datetime=None):
        """average impact aggregated over all exposures

        Parameters
        ----------
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.
        Returns
        -------
        float
        """
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        return self._impact[haz_ind].aai_agg

    def haz_summary_str(self, run_datetime=None):
        """provide a summary string for the hazard part of the forecast

        Parameters
        ----------
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.
        Returns
        -------
        str
            summarizing the most important information about the hazard
        """
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        return (
            self.hazard[haz_ind].tag.haz_type
            + "_"
            + self.haz_model
            + "_run"
            + run_datetime.strftime("%Y%m%d%H")
            + "_event"
            + self.event_date.strftime("%Y%m%d")
        )

    def summary_str(self, run_datetime=None):
        """provide a summary string for the impact forecast

        Parameters
        ----------
        run_datetime: datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.

        Returns
        -------
        str
            summarizing the most important information about
            the impact forecast
        """
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        return self.haz_summary_str(run_datetime) + "_" + self.exposure_name

    def lead_time(self, run_datetime=None):
        """provide the lead time for the impact forecast

        Parameters
        ----------
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.

        Returns
        -------
        datetime.timedelta
            the difference between the initialization time of the forecast
            model run and the date of the event, commenly named lead time
        """
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        return self.event_date - run_datetime

    def calc(self, force_reassign=False):
        """calculate the impacts for all lead times using
        exposure, all hazards of all run_datetime, and ImpactFunctionSet.

        Parameters
        ----------
        force_reassign : bool, optional
            Reassign hazard centroids to the exposure for all hazards,
            default is false.
        """
        # calc impact
        for ind_i, haz_i in enumerate(self.hazard):
            # force reassign
            if force_reassign:
                self.exposure.assign_centroids(haz_i)
            self._impact[ind_i].calc(
                self.exposure, self.vulnerability, haz_i, save_mat=True
            )

    def plot_imp_map(
        self,
        run_datetime=None,
        save_fig=True,
        close_fig=False,
        polygon_file=None,
        polygon_file_crs="epsg:4326",
        proj=ccrs.PlateCarree(),
        figsize=(9, 13),
        adapt_fontsize=True,
    ):
        """plot a map of the impacts

        Parameters
        ----------
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.
        save_fig : bool, optional
            Figure is saved if True, folder is within your configurable
            save_dir and filename is derived from the method summary_str()
            (for more details see class docstring).  Default is True.
        close_fig : bool, optional
            Figure not drawn if True. Default is False.
        polygon_file : str, optional
            Points to a .shp-file with polygons do be drawn as outlines on
            the plot, default is None to not draw the lines. please also
            specify the crs in the parameter polygon_file_crs.
        polygon_file_crs : str, optional
            String of pattern <provider>:<code> specifying
            the crs. has to be readable by pyproj.Proj. Default is
            'epsg:4326'.
        proj : ccrs
            coordinate reference system used in coordinates
            The default is ccrs.PlateCarree()
        figsize : tuple
            figure size for plt.subplots, width, height in inches
            The default is (9, 13)
        adapt_fontsize : bool, optional
            If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
            the default matplotlib font size is used. Default is True.

        Returns
        -------
        axes: cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        # select hazard with run_datetime
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        # plot impact map of all runs
        map_file_name = self.summary_str(run_datetime) + "_impact_map" + ".jpeg"
        map_file_name_full = FORECAST_PLOT_DIR / map_file_name
        lead_time_str = "{:.0f}".format(
            self.lead_time(run_datetime).days
            + self.lead_time(run_datetime).seconds / 60 / 60 / 24
        )
        title_dict = {
            "event_day": self.event_date.strftime("%a %d %b %Y 00-24UTC"),
            "run_start": (
                run_datetime.strftime("%d.%m.%Y %HUTC +") + lead_time_str + "d"
            ),
            "explain_text": ("mean building damage caused by wind"),
            "model_text": "CLIMADA IMPACT",
        }
        fig, axes = self._plot_imp_map(
            run_datetime,
            title=title_dict,
            cbar_label=(
                "forecasted damage per gridcell [" + self._impact[haz_ind].unit + "]"
            ),
            polygon_file=polygon_file,
            polygon_file_crs=polygon_file_crs,
            proj=proj,
            figsize=figsize,
            adapt_fontsize=adapt_fontsize,
        )
        if save_fig:
            fig.savefig(map_file_name_full)
        if close_fig:
            fig.clf()
            plt.close(fig)
        return axes

    def _plot_imp_map(
        self,
        run_datetime,
        title,
        cbar_label,
        polygon_file=None,
        polygon_file_crs="epsg:4326",
        proj=ccrs.PlateCarree(),
        figsize=(9, 13),
        adapt_fontsize=True,
    ):
        # select hazard with run_datetime
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        # tryout new plot with right projection
        extend = "neither"
        value = self._impact[haz_ind].eai_exp
        #    value[np.invert(mask)] = np.nan
        coord = self._impact[haz_ind].coord_exp

        # Generate array of values used in each subplot
        array_sub = value
        shapes = True
        if not polygon_file:
            shapes = False
        var_name = cbar_label
        geo_coord = coord
        num_im, list_arr = u_plot._get_collection_arrays(array_sub)
        list_tit = to_list(num_im, title, "title")
        list_name = to_list(num_im, var_name, "var_name")
        list_coord = to_list(num_im, geo_coord, "geo_coord")

        kwargs = dict()
        kwargs["cmap"] = CMAP_IMPACT
        kwargs["s"] = 5
        kwargs["marker"] = ","
        kwargs["norm"] = BoundaryNorm(
            np.append(
                np.append([0], [10**x for x in np.arange(0, 2.9, 2.9 / 9)]),
                [10**x for x in np.arange(3, 7, 4 / 90)],
            ),
            CMAP_IMPACT.N,
            clip=True,
        )

        # Generate each subplot
        fig, axis_sub, _fontsize = u_plot.make_map(
            num_im, proj=proj, figsize=figsize, adapt_fontsize=adapt_fontsize
        )
        if not isinstance(axis_sub, np.ndarray):
            axis_sub = np.array([[axis_sub]])
        fig.set_size_inches(9, 8)
        for array_im, axis, tit, name, coord in zip(
            list_arr, axis_sub.flatten(), list_tit, list_name, list_coord
        ):
            if coord.shape[0] != array_im.size:
                raise ValueError(
                    "Size mismatch in input array: %s != %s."
                    % (coord.shape[0], array_im.size)
                )
            # Binned image with coastlines
            extent = u_plot._get_borders(coord)
            axis.set_extent((extent), ccrs.PlateCarree())
            hex_bin = axis.scatter(
                coord[:, 1],
                coord[:, 0],
                c=array_im,
                transform=ccrs.PlateCarree(),
                **kwargs
            )
            if shapes:
                # add warning regions
                shp = shapereader.Reader(polygon_file)
                transformer = pyproj.Transformer.from_crs(
                    polygon_file_crs, self._impact[haz_ind].crs, always_xy=True
                )
                for geometry, _ in zip(shp.geometries(), shp.records()):
                    geom2 = shapely.ops.transform(transformer.transform, geometry)
                    axis.add_geometries(
                        [geom2],
                        crs=ccrs.PlateCarree(),
                        facecolor="none",
                        edgecolor="gray",
                    )
            else:  # add country boundaries
                u_plot.add_shapes(axis)
            # Create colorbar in this axis
            cbax = make_axes_locatable(axis).append_axes(
                "bottom", size="6.5%", pad=0.3, axes_class=plt.Axes
            )
            cbar = plt.colorbar(
                hex_bin, cax=cbax, orientation="horizontal", extend=extend
            )
            cbar.set_label(name)
            cbar.formatter.set_scientific(False)
            cbar.set_ticks([0, 1000, 10000, 100000, 1000000])
            cbar.set_ticklabels(["0", "1 000", "10 000", "100 000", "1 000 000"])
            title_position = {
                "model_text": [0.02, 0.85],
                "explain_text": [0.02, 0.81],
                "event_day": [0.98, 0.85],
                "run_start": [0.98, 0.81],
            }
            left_right = {
                "model_text": "left",
                "explain_text": "left",
                "event_day": "right",
                "run_start": "right",
            }
            color = {
                "model_text": "k",
                "explain_text": "k",
                "event_day": "r",
                "run_start": "k",
            }
            for t_i in tit:
                plt.figtext(
                    title_position[t_i][0],
                    title_position[t_i][1],
                    tit[t_i],
                    fontsize="xx-large",
                    color=color[t_i],
                    ha=left_right[t_i],
                )

        fig.tight_layout()
        fig.subplots_adjust(top=0.8)
        return fig, axis_sub

    def plot_hist(
        self, run_datetime=None, save_fig=True, close_fig=False, figsize=(9, 8)
    ):
        """plot histogram of the forecasted impacts all ensemble members

        Parameters
        ----------
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.
        save_fig : bool, optional
            Figure is saved if True, folder is within your configurable
            save_dir and filename is derived from the method summary_str()
            (for more details see class docstring).  Default is True.
        close_fig : bool, optional
            Figure is not drawn if True. Default is False.
        figsize : tuple
            figure size for plt.subplots, width, height in inches
            The default is (9, 8)
        Returns
        -------
        axes : matplotlib.axes.Axes
        """
        # select hazard with run_datetime
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        # plot histogram of all runs
        histbin_file_name = self.summary_str(run_datetime) + "_histbin" + ".svg"
        histbin_file_name_full = FORECAST_PLOT_DIR / histbin_file_name
        lower_bound = np.max(
            [
                np.floor(
                    np.log10(np.max([np.min(self._impact[haz_ind].at_event), 0.1]))
                ),
                0,
            ]
        )
        upper_bound = (
            np.max(
                [
                    np.ceil(
                        np.log10(np.max([np.max(self._impact[haz_ind].at_event), 0.1]))
                    ),
                    lower_bound + 5,
                ]
            )
            + 0.1
        )
        bins_log = np.arange(lower_bound, upper_bound, 0.5)
        bins = 10**bins_log
        fig = plt.figure(figsize=figsize)
        axes = fig.add_subplot(111)
        plt.xscale("log")
        plt.hist(
            [np.max([x, 1]) for x in self._impact[haz_ind].at_event],
            bins=bins,
            weights=np.ones(len(self._impact[haz_ind].at_event))
            / len(self._impact[haz_ind].at_event),
        )
        axes.yaxis.set_major_formatter(PercentFormatter(1))
        formatter = ScalarFormatter()
        formatter.set_scientific(False)  # formatter.format = "%:'.0f"
        axes.xaxis.set_major_formatter(formatter)

        x_ticklabels = list()
        x_ticks = list()
        ticklabel_dict = {
            pow(10, i): self._number_to_str(pow(10, i)) for i in range(0, 15)
        }  # turn int to str

        for i in np.arange(15):
            tick_i = 10.0**i
            x_ticks.append(tick_i)
            x_ticklabels.append(ticklabel_dict[tick_i])
        axes.xaxis.set_ticks(x_ticks)
        axes.xaxis.set_ticklabels(x_ticklabels)
        plt.xticks(rotation=15, horizontalalignment="right")
        plt.xlim([(10**-0.25) * bins[0], (10**0.25) * bins[-1]])

        lead_time_str = "{:.0f}".format(
            self.lead_time(run_datetime).days
            + self.lead_time(run_datetime).seconds / 60 / 60 / 24
        )
        title_dict = {
            "event_day": self.event_date.strftime("%a %d %b %Y 00-24UTC"),
            "run_start": (
                run_datetime.strftime("%d.%m.%Y %HUTC +") + lead_time_str + "d"
            ),
            "explain_text": ("total building damage"),
            "model_text": "CLIMADA IMPACT",
        }
        title_position = {
            "model_text": [0.13, 0.94],
            "explain_text": [0.13, 0.9],
            "event_day": [0.9, 0.94],
            "run_start": [0.9, 0.9],
        }
        left_right = {
            "model_text": "left",
            "explain_text": "left",
            "event_day": "right",
            "run_start": "right",
        }
        color = {
            "model_text": "k",
            "explain_text": "k",
            "event_day": "r",
            "run_start": "k",
        }
        for t_i in title_dict:
            plt.figtext(
                title_position[t_i][0],
                title_position[t_i][1],
                title_dict[t_i],
                fontsize="xx-large",
                color=color[t_i],
                ha=left_right[t_i],
            )

        plt.xlabel(
            "forecasted total damage for "
            + self.exposure_name
            + " ["
            + self._impact[haz_ind].unit
            + "]"
        )
        plt.ylabel("probability")
        plt.text(
            0.75,
            0.85,
            "mean damage:\nCHF "
            + self._number_to_str(self._impact[haz_ind].at_event.mean()),
            horizontalalignment="center",
            verticalalignment="center",
            transform=axes.transAxes,
        )
        if save_fig:
            plt.savefig(histbin_file_name_full)
        if close_fig:
            plt.clf()
            plt.close(fig)
        return axes

    @staticmethod
    def _number_to_str(number):
        """Generate numbers as str (with thousand, million, trillion).
            Example: 1000 becomes 1 thousand

        Parameters
        ----------
        number : int

        Returns
        -------
        str
        """
        number_names = {
            1: "",
            1000: "thousand",
            1000000: "million",
            1000000000: "billion",
            1000000000000: "trillion",
        }
        [value], name = u_value_to_monetary_unit(number, abbreviations=number_names)
        return "{:.0f} {}".format(value, name)

    def plot_exceedence_prob(
        self,
        threshold,
        explain_str=None,
        run_datetime=None,
        save_fig=True,
        close_fig=False,
        polygon_file=None,
        polygon_file_crs="epsg:4326",
        proj=ccrs.PlateCarree(),
        figsize=(9, 13),
        adapt_fontsize=True,
    ):
        """plot exceedence map

        Parameters
        ----------
        threshold : float
            Threshold of impact unit for which exceedence probability
            should be plotted.
        explain_str : str, optional
            Short str which explains threshold, explain_str is included
            in the title of the figure.
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.
        save_fig : bool, optional
            Figure is saved if True, folder is within your configurable
            save_dir and filename is derived from the method summary_str()
            (for more details see class docstring).  Default is True.
        close_fig : bool, optional
            Figure not drawn if True. Default is False.
        polygon_file : str, optional
            Points to a .shp-file with polygons do be drawn as outlines on
            the plot, default is None to not draw the lines. please also
            specify the crs in the parameter polygon_file_crs.
        polygon_file_crs : str, optional
            String of pattern <provider>:<code> specifying
            the crs. has to be readable by pyproj.Proj. Default is
            'epsg:4326'.
        proj : ccrs
            coordinate reference system used in coordinates
            The default is ccrs.PlateCarree()
        figsize : tuple
            figure size for plt.subplots, width, height in inches
            The default is (9, 13)
        adapt_fontsize : bool, optional
            If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
            the default matplotlib font size is used. Default is True.

        Returns
        -------
        axes: cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        # select hazard with run_datetime
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        wind_map_file_name = (
            self.summary_str(run_datetime) + "_exceed_" + str(threshold) + "_map.jpeg"
        )
        wind_map_file_name_full = FORECAST_PLOT_DIR / wind_map_file_name
        lead_time_str = "{:.0f}".format(
            self.lead_time(run_datetime).days
            + self.lead_time(run_datetime).seconds / 60 / 60 / 24
        )
        title_dict = {
            "event_day": self.event_date.strftime("%a %d %b %Y 00-24UTC"),
            "run_start": (
                run_datetime.strftime("%d.%m.%Y %HUTC +") + lead_time_str + "d"
            ),
            "explain_text": (
                "threshold: " + str(threshold) + " " + self._impact[haz_ind].unit
            )
            if explain_str is None
            else explain_str,
            "model_text": "Exceedance probability map",
        }
        cbar_label = "probabilty of reaching threshold"
        fig, axes = self._plot_exc_prob(
            run_datetime,
            threshold,
            title_dict,
            cbar_label,
            proj,
            polygon_file=polygon_file,
            polygon_file_crs=polygon_file_crs,
            figsize=figsize,
            adapt_fontsize=adapt_fontsize,
        )
        if save_fig:
            plt.savefig(wind_map_file_name_full)
        if close_fig:
            plt.clf()
            plt.close(fig)
        return axes

    def _plot_exc_prob(
        self,
        run_datetime,
        threshold,
        title,
        cbar_label,
        proj=ccrs.PlateCarree(),
        polygon_file=None,
        polygon_file_crs="epsg:4326",
        mask=None,
        figsize=(9, 13),
        adapt_fontsize=True,
    ):
        """plot the probability of reaching a threshold"""
        # select hazard with run_datetime
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        extend = "neither"
        value = np.squeeze(
            np.asarray(
                (self._impact[haz_ind].imp_mat > threshold).sum(axis=0)
                / self._impact[haz_ind].event_id.size
            )
        )
        if mask is not None:
            value[np.invert(mask)] = np.nan
        coord = self._impact[haz_ind].coord_exp
        # Generate array of values used in each subplot
        array_sub = value
        shapes = True
        if not polygon_file:
            shapes = False
        var_name = cbar_label
        geo_coord = coord
        num_im, list_arr = u_plot._get_collection_arrays(array_sub)
        list_tit = to_list(num_im, title, "title")
        list_name = to_list(num_im, var_name, "var_name")
        list_coord = to_list(num_im, geo_coord, "geo_coord")

        kwargs = dict()
        kwargs["cmap"] = CMAP_WARNPROB
        kwargs["s"] = 5
        kwargs["marker"] = ","
        kwargs["norm"] = BoundaryNorm(np.linspace(0, 1, 11), CMAP_WARNPROB.N, clip=True)

        # Generate each subplot
        fig, axis_sub, _fontsize = u_plot.make_map(
            num_im, proj=proj, figsize=figsize, adapt_fontsize=adapt_fontsize
        )
        if not isinstance(axis_sub, np.ndarray):
            axis_sub = np.array([[axis_sub]])
        fig.set_size_inches(9, 8)
        for array_im, axis, tit, name, coord in zip(
            list_arr, axis_sub.flatten(), list_tit, list_name, list_coord
        ):
            if coord.shape[0] != array_im.size:
                raise ValueError(
                    "Size mismatch in input array: %s != %s."
                    % (coord.shape[0], array_im.size)
                )

            hex_bin = axis.scatter(
                coord[:, 1],
                coord[:, 0],
                c=array_im,
                transform=ccrs.PlateCarree(),
                **kwargs
            )
            if shapes:
                # add warning regions
                shp = shapereader.Reader(polygon_file)
                transformer = pyproj.Transformer.from_crs(
                    polygon_file_crs, self._impact[haz_ind].crs, always_xy=True
                )
                for geometry, _ in zip(shp.geometries(), shp.records()):
                    geom2 = shapely.ops.transform(transformer.transform, geometry)
                    axis.add_geometries(
                        [geom2],
                        crs=ccrs.PlateCarree(),
                        facecolor="none",
                        edgecolor="gray",
                    )

            # Create colorbar in this axis
            cbax = make_axes_locatable(axis).append_axes(
                "bottom", size="6.5%", pad=0.3, axes_class=plt.Axes
            )
            cbar = plt.colorbar(
                hex_bin, cax=cbax, orientation="horizontal", extend=extend
            )
            cbar.set_label(name)
            title_position = {
                "model_text": [0.02, 0.94],
                "explain_text": [0.02, 0.9],
                "event_day": [0.98, 0.94],
                "run_start": [0.98, 0.9],
            }
            left_right = {
                "model_text": "left",
                "explain_text": "left",
                "event_day": "right",
                "run_start": "right",
            }
            color = {
                "model_text": "k",
                "explain_text": "k",
                "event_day": "r",
                "run_start": "k",
            }
            for t_i in tit:
                plt.figtext(
                    title_position[t_i][0],
                    title_position[t_i][1],
                    tit[t_i],
                    fontsize="xx-large",
                    color=color[t_i],
                    ha=left_right[t_i],
                )
            extent = u_plot._get_borders(coord)
            axis.set_extent((extent), ccrs.PlateCarree())
        fig.tight_layout()
        return fig, axis_sub

    def plot_warn_map(
        self,
        polygon_file=None,
        polygon_file_crs="epsg:4326",
        thresholds="default",
        decision_level="exposure_point",
        probability_aggregation=0.5,
        area_aggregation=0.5,
        title="WARNINGS",
        explain_text="warn level based on thresholds",
        run_datetime=None,
        proj=ccrs.PlateCarree(),
        figsize=(9, 13),
        save_fig=True,
        close_fig=False,
        adapt_fontsize=True,
    ):
        """plot map colored with 5 warning colors for all regions in provided
        shape file.

        Parameters
        ----------
        polygon_file : str, optional
            path to shp-file containing warning region polygons
        polygon_file_crs : str, optional
            String of pattern <provider>:<code> specifying
            the crs. has to be readable by pyproj.Proj. Default is
            'epsg:4326'.
        thresholds : list of 4 floats, optional
            Thresholds for coloring region in second, third, forth
            and fifth warning color.
        decision_level : str, optional
            Either 'exposure_point'  or 'polygon'. Default value is
            'exposure_point'.
        probability_aggregation : float or str, optional
            Either a float between [0..1] spezifying a quantile
            or 'mean' or 'sum'. Default value is 0.5.
        area_aggregation : float or str.
            Either a float between [0..1] specifying a quantile
            or 'mean' or 'sum'. Default value is 0.5.
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.
        title : str, optional
            Default is 'WARNINGS'.
        explain_text : str, optional
            Defaut is 'warn level based on thresholds'.
        proj : ccrs
            coordinate reference system used in coordinates
        figsize : tuple
            figure size for plt.subplots, width, height in inches
            The default is (9, 13)
        save_fig : bool, optional
            Figure is saved if True, folder is within your configurable
            save_dir and filename is derived from the method summary_str()
            (for more details see class docstring).  Default is True.
        close_fig : bool, optional
            Figure is not drawn if True. The default is False.
        adapt_fontsize : bool, optional
            If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
            the default matplotlib font size is used. Default is True.

        Returns
        -------
        axes : cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        # select hazard with run_datetime
        if thresholds == "default":
            thresholds = [2, 3, 4, 5]
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        warn_map_file_name = self.summary_str(run_datetime) + "_warn_map.jpeg"
        warn_map_file_name_full = FORECAST_PLOT_DIR / warn_map_file_name
        decision_dict = {
            "probability_aggregation": probability_aggregation,
            "area_aggregation": area_aggregation,
        }
        lead_time_str = "{:.0f}".format(
            self.lead_time(run_datetime).days
            + self.lead_time(run_datetime).seconds / 60 / 60 / 24
        )
        title_dict = {
            "event_day": self.event_date.strftime("%a %d %b %Y 00-24UTC"),
            "run_start": (
                run_datetime.strftime("%d.%m.%Y %HUTC +") + lead_time_str + "d"
            ),
            "explain_text": explain_text,
            "model_text": title,
        }

        fig, axes = self._plot_warn(
            run_datetime,
            thresholds,
            decision_level,
            decision_dict,
            polygon_file,
            polygon_file_crs,
            title_dict,
            proj,
            figsize=figsize,
            adapt_fontsize=adapt_fontsize,
        )
        if save_fig:
            plt.savefig(warn_map_file_name_full)
        if close_fig:
            plt.clf()
            plt.close(fig)
        return axes

    def _plot_warn(
        self,
        run_datetime,
        thresholds,
        decision_level,
        decision_dict,
        polygon_file,
        polygon_file_crs,
        title,
        proj=ccrs.PlateCarree(),
        figsize=(9, 13),
        adapt_fontsize=True,
    ):
        """plotting the warning level of each warning region based on thresholds"""
        # select hazard with run_datetime
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]

        kwargs = dict()
        kwargs["cmap"] = CMAP_WARNPROB
        kwargs["s"] = 5
        kwargs["marker"] = ","
        kwargs["norm"] = BoundaryNorm(np.linspace(0, 1, 11), CMAP_WARNPROB.N, clip=True)

        # Generate each subplot
        fig, axis, _fontsize = u_plot.make_map(
            1, proj=proj, figsize=figsize, adapt_fontsize=adapt_fontsize
        )
        if isinstance(axis, np.ndarray):
            axis = axis[0]
        tit = title
        fig.set_size_inches(9, 8)

        # add warning regions
        shp = shapereader.Reader(polygon_file)
        transformer = pyproj.Transformer.from_crs(
            polygon_file_crs, self._impact[haz_ind].crs, always_xy=True
        )
        # checking the decision dict and define the corresponding functions
        if not (
            isinstance(decision_dict["probability_aggregation"], float)
            & isinstance(decision_dict["area_aggregation"], float)
        ):
            ValueError(
                " If decision_level is 'exposure_point',"
                + "parameters probability_aggregation and "
                + "area_aggregation of "
                + "Forecast.plot_warn_map() must both be "
                + "floats between [0..1]. Which each "
                + "specify quantiles."
            )
        decision_dict_functions = decision_dict.copy()
        for aggregation in decision_dict:
            if isinstance(decision_dict[aggregation], float):
                decision_dict_functions[aggregation] = np.percentile
            elif decision_dict[aggregation] == "sum":
                decision_dict_functions[aggregation] = np.sum
            elif decision_dict[aggregation] == "mean":
                decision_dict_functions[aggregation] = np.mean
            else:
                raise ValueError(
                    "Parameter area_aggregation of "
                    + "Forecast.plot_warn_map() must eiter be "
                    + "a float between [0..1], which "
                    + "specifys a quantile. or 'sum' or 'mean'."
                )

        for geometry, _ in zip(shp.geometries(), shp.records()):
            geom2 = shapely.ops.transform(transformer.transform, geometry)
            in_geom = u_coord.coord_on_land(
                lat=self._impact[haz_ind].coord_exp[:, 0],
                lon=self._impact[haz_ind].coord_exp[:, 1],
                land_geom=geom2,
            )
            if not in_geom.any():
                continue
            # decide warning level
            warn_level = 0
            for ind_i, warn_thres_i in enumerate(thresholds):
                if decision_level == "exposure_point":
                    # decision at each grid_point
                    probabilities = np.squeeze(
                        np.asarray(
                            (self._impact[haz_ind].imp_mat >= warn_thres_i).sum(axis=0)
                            / self._impact[haz_ind].event_id.size
                        )
                    )
                    # quantiles over probability
                    area = (
                        probabilities[in_geom]
                        >= decision_dict["probability_aggregation"]
                    ).sum()
                    # quantiles over area
                    if area >= (in_geom.sum() * decision_dict["area_aggregation"]):
                        warn_level = ind_i + 1
                elif decision_level == "polygon":
                    # aggregation over area
                    if isinstance(decision_dict["area_aggregation"], float):
                        value_per_member = decision_dict_functions["area_aggregation"](
                            self._impact[haz_ind].imp_mat[:, in_geom].todense(),
                            decision_dict["area_aggregation"],
                            axis=1,
                        )
                    else:
                        value_per_member = decision_dict_functions["area_aggregation"](
                            self._impact[haz_ind].imp_mat[:, in_geom].todense(), axis=1
                        )
                    # aggregation over members/probability
                    if isinstance(decision_dict["probability_aggregation"], float):
                        value_per_region = decision_dict_functions[
                            "probability_aggregation"
                        ](value_per_member, decision_dict["probability_aggregation"])
                    else:
                        value_per_region = decision_dict_functions[
                            "probability_aggregation"
                        ](value_per_member)
                    # warn level decision
                    if value_per_region >= warn_thres_i:
                        warn_level = ind_i + 1
                else:
                    raise ValueError(
                        "Parameter decision_level of "
                        + "Forecast.plot_warn_map() must eiter be "
                        + "'exposure_point' or 'polygon'."
                    )
            # plot warn_region with specific color (dependent on warning level)
            axis.add_geometries(
                [geom2],
                crs=ccrs.PlateCarree(),
                facecolor=COLORS_WARN[warn_level, :],
                edgecolor="gray",
            )

        # Create legend in this axis
        hazard_levels = [
            "1: Minimal or no hazard",
            "2: Moderate hazard",
            "3: Significant hazard",
            "4: Severe hazard",
            "5: Very severe hazard",
        ]
        legend_elements = [
            Patch(facecolor=COLORS_WARN[n, :], edgecolor="gray", label=hazard_level)
            for n, hazard_level in enumerate(hazard_levels)
        ]

        axis.legend(
            handles=legend_elements,
            loc="upper center",
            framealpha=0.5,
            bbox_to_anchor=(0.5, -0.02),
            ncol=3,
        )
        title_position = {
            "model_text": [0.02, 0.91],
            "explain_text": [0.02, 0.87],
            "event_day": [0.98, 0.91],
            "run_start": [0.98, 0.87],
        }
        left_right = {
            "model_text": "left",
            "explain_text": "left",
            "event_day": "right",
            "run_start": "right",
        }
        color = {
            "model_text": "k",
            "explain_text": "k",
            "event_day": "r",
            "run_start": "k",
        }
        for t_i in tit:
            plt.figtext(
                title_position[t_i][0],
                title_position[t_i][1],
                tit[t_i],
                fontsize="xx-large",
                color=color[t_i],
                ha=left_right[t_i],
            )

        extent = u_plot._get_borders(self._impact[haz_ind].coord_exp)
        axis.set_extent((extent), ccrs.PlateCarree())
        fig.tight_layout()
        return fig, axis

    def plot_hexbin_ei_exposure(self, run_datetime=None, figsize=(9, 13)):
        """plot the expected impact

        Parameters
        ----------
        run_datetime : datetime.datetime, optional
            Select the used hazard by the run_datetime,
            default is first element of attribute run_datetime.
        figsize : tuple
            figure size for plt.subplots, width, height in inches
            The default is (9, 13)
        Returns
        -------
        axes : cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        # select hazard with run_datetime
        if run_datetime is None:
            run_datetime = self.run_datetime[0]
        haz_ind = np.argwhere(np.isin(self.run_datetime, run_datetime))[0][0]
        return self._impact[haz_ind].plot_hexbin_eai_exposure(figsize=figsize)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test of util.math module
"""


from climada.util.value_representation import sig_dig, sig_dig_list, ABBREV
from climada.util.value_representation import value_to_monetary_unit
import unittest
import numpy as np
import math


class TestDigits(unittest.TestCase):
    """Test date functions"""

    def test_sig_dig_pass(self):
        """Test sig_dig function"""
        n_sig_dig = 3
        nbs_in = [1.2345, 12345, -12345, -12.345]
        nbs_out = [1.23, 12300, -12300, -12.3]
        for nb_in, nb_out in zip(nbs_in, nbs_out):
            self.assertEqual(sig_dig(nb_in, n_sig_dig), nb_out)
        self.assertTrue(
            np.array_equal(sig_dig_list(nbs_in, n_sig_dig), nbs_out)
            )

    def test_sig_dig_fail(self):
        """Test sig_dig function"""
        n_sig_dig_wrong = 4
        nbs_in = [1.2345, 12345, -12345, -12.345]
        nbs_out = [1.23, 12300, -12300, -12.3]
        for nb_in, nb_out in zip(nbs_in, nbs_out):
            self.assertNotEqual(sig_dig(nb_in, n_sig_dig_wrong), nb_out)
        self.assertFalse(
            np.array_equal(sig_dig_list(nbs_in, n_sig_dig_wrong), nbs_out)
            )

    def test_value_to_monetary_unit_pass(self):
        """Test money_unit function"""
        nbs_in = [-1e10, -1e6, -1e2, 0, 1e3, 1e7, 1e11]
        nbs_out = [-10, -1, -100, 0, 1, 10, 100]
        names_out = [ABBREV[1e9], ABBREV[1e6], ABBREV[1], ABBREV[1],
                 ABBREV[1e3], ABBREV[1e6], ABBREV[1e9]]
        for j, (nb_in, nb_out) in enumerate(zip(nbs_in, nbs_out)):
            money, names = value_to_monetary_unit(nb_in)
            self.assertEqual(money[0], nb_out)
            self.assertEqual(names, names_out[j])

    def test_value_to_monetary_unit_0inf_pass(self):
        """Test money_unit function"""
        nbs_in = [-math.inf, 0, 1e-10, 1e-5, math.inf]
        nbs_out = [-math.inf, 0, 1e-10, 1e-5, math.inf]
        names_out = [ABBREV[1], ABBREV[1], ABBREV[1], ABBREV[1],
                 ABBREV[1], ABBREV[1], ABBREV[1]]
        for j, (nb_in, nb_out) in enumerate(zip(nbs_in, nbs_out)):
            money, names = value_to_monetary_unit(nb_in)
            self.assertEqual(money[0], nb_out)
            self.assertEqual(names, names_out[j])

    def test_value_to_monetary_unit_nan_pass(self):
        """Test money_unit function"""
        nb_in = math.nan
        money, name = value_to_monetary_unit(nb_in)
        self.assertTrue(math.isnan(money[0]))
        self.assertEqual(name, '')


    def test_value_to_monetary_unit_sigdig_pass(self):
        """Test money_unit function with significant digits"""
        nbs_in = [-1e10*1.2345, -1e6*1.2345, -1e2*1.2345, 0, 1e3*1.2345,
                  1e7*1.2345, 1e11*1.2345]
        nbs_out = [-12.3, -1.23, -123, 0, 1.23, 12.3, 123]
        names_out = [ABBREV[1e9], ABBREV[1e6], ABBREV[1], ABBREV[1],
                 ABBREV[1e3], ABBREV[1e6], ABBREV[1e9]]
        for j, (nb_in, nb_out) in enumerate(zip(nbs_in, nbs_out)):
            money, names = value_to_monetary_unit(nb_in, n_sig_dig=3)
            self.assertEqual(money[0], nb_out)
            self.assertEqual(names, names_out[j])

    def test_value_to_monetary_unit_list_pass(self):
        """Test money_unit function with list of numbers"""
        nbs_in = [-1e10*1.2345, -1e9*1.2345]
        nbs_out = [-12.3, -1.23]
        name_out = ABBREV[1e9]
        money, name = value_to_monetary_unit(nbs_in, n_sig_dig=3)
        self.assertTrue(np.array_equal(money, nbs_out))
        self.assertEqual(name, name_out)
        nbs_in = [1e4*1.2345, 1e3*1.2345, 1e2*1.2345,]
        nbs_out = [12.3, 1.23, 0.123]
        name_out = ABBREV[1e3]
        money, name = value_to_monetary_unit(nbs_in, n_sig_dig=3)
        self.assertTrue(np.array_equal(money, nbs_out))
        self.assertEqual(name, name_out)

    def test_value_to_monetary_unit_list_0inf_pass(self):
        """Test money_unit function with list of numbers"""
        nbs_in = [-1e10*1.2345, -1e9*1.2345, 0, math.inf]
        nbs_out = [-12.3, -1.23, 0, math.inf]
        name_out = ABBREV[1e9]
        money, name = value_to_monetary_unit(nbs_in, n_sig_dig=3)
        self.assertTrue(np.array_equal(money, nbs_out))
        self.assertEqual(name, name_out)
        nbs_in = [1e4*1.2345, 1e3*1.2345, 1e2*1.2345, 0, math.inf]
        nbs_out = [12.3, 1.23, 0.123, 0, math.inf]
        name_out = ABBREV[1e3]
        money, name = value_to_monetary_unit(nbs_in, n_sig_dig=3)
        self.assertTrue(np.array_equal(money, nbs_out))
        self.assertEqual(name, name_out)

    def test_value_to_monetary_unit_list_nan_pass(self):
        """Test money_unit function with list of numbers"""
        nbs_in = [-1e10*1.2345, -1e9*1.2345, math.nan]
        nbs_out = [-12.3, -1.23, math.nan]
        name_out = ABBREV[1e9]
        money, name = value_to_monetary_unit(nbs_in, n_sig_dig=3)
        self.assertTrue(math.isnan(nbs_out[-1]))
        self.assertTrue(np.array_equal(money[:-1], nbs_out[:-1]))
        self.assertEqual(name, name_out)


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestDigits)
    unittest.TextTestRunner(verbosity=2).run(TESTS)
